#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: MBI

  Description: @{shortdesc}@
    @{longdesc}@

  Version of MPI: Conforms to MPI 1.1, does not require MPI 2 implementation

BEGIN_MPI_FEATURES
  P2P!basic: @{p2pfeature}@
  P2P!nonblocking: @{ip2pfeature}@
  P2P!persistent: @{persfeature}0@
  COLL!basic: Lacking
  COLL!nonblocking: Lacking
  COLL!persistent: Lacking
  COLL!tools: Lacking
  RMA: Lacking
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 2 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

#define buff_size 1

int main(int argc, char **argv) {
  int nprocs = -1;
  int rank = -1;
  int its_raining = 0;
  int src=0, dest=1;
  int stag=0, rtag=0;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (nprocs < 2)
    printf("MBI ERROR: This test needs at least 2 processes to produce a bug!\\n");

  MPI_Comm newcom = MPI_COMM_WORLD;
  MPI_Datatype type = MPI_INT;

  @{init1}@
  @{init2}@
  if (rank == 0) {
    @{operation1}@ /* MBIERROR1 */
    @{fini1}@
  }else if (@{change_cond}@){
    @{operation2}@ /* MBIERROR2 */
    @{fini2}@
  }

  MPI_Finalize();
  printf("Rank %d finished normally\\n", rank);
  return 0;
}
"""


for p in gen.send + gen.ssend + gen.bsend + gen.recv + gen.irecv + gen.isend:
    patterns = {}
    patterns = {'p': p}
    patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
    patterns['p2pfeature'] = 'Yes' if p in gen.send + gen.bsend + gen.ssend + gen.recv else 'Lacking'
    patterns['ip2pfeature'] = 'Yes' if p in gen.isend + gen.irecv else 'Lacking'
    patterns['persfeature'] = 'Lacking'
    #patterns['persfeature'] = 'Yes' if p in psend + precv else 'Lacking'
    patterns['p'] = p
    patterns['init1'] = gen.init[p]("1")
    patterns['init2'] = '' #gen.init[p2]("2")
    patterns['fini1'] = gen.fini[p]("1")
    patterns['fini2'] = '' #gen.fini[p2]("2")
    patterns['operation1'] = gen.operation[p]("1")
    patterns['operation2'] = '' #gen.operation[p2]("2")
    patterns['change_cond'] = 'rank == 1'

    # Generate the incorrect matching with one call
    replace = patterns.copy()
    replace['shortdesc'] = 'Point to point @{p}@ is not matched'
    replace['longdesc'] = 'Process 0 calls @{p}@ and is not matched'
    replace['outcome'] = 'ERROR: CallMatching'
    replace['errormsg'] = 'P2P mistmatch. @{p}@ at @{filename}@:@{line:MBIERROR1}@ is not matched.'
    gen.make_file(template, f'CallOrdering_{p}_nok.c', replace)

    # Generate the incorrect matching with two calls
    replace = patterns.copy()
    replace['shortdesc'] = 'Both point to point @{p}@ are not matched'
    replace['longdesc'] = 'Processes 0 and 1 both call @{p}@ which are not matched'
    replace['outcome'] = 'ERROR: CallMatching'
    replace['errormsg'] = 'P2P mismatch. @{p}@ at @{filename}@:@{line:MBIERROR1}@ and @{p}@ at @{filename}@:@{line:MBIERROR2}@ are not matched.'
    replace['operation2'] = gen.operation[p]("1")
    replace['fini2'] = gen.fini[p]("1")
    gen.make_file(template, f'CallOrdering_{p}_{p}_nok.c', replace)

for s in gen.send + gen.isend + gen.ssend + gen.bsend:
    for r in gen.recv + gen.irecv:
        patterns = {}
        patterns = {'s': s, 'r': r}
        patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
        patterns['p2pfeature'] = 'Yes' if s in gen.send or r in gen.recv else 'Lacking'
        patterns['ip2pfeature'] = 'Yes' if s in gen.isend or r in gen.irecv else 'Lacking'
        patterns['s'] = s
        patterns['r'] = r
        patterns['init1'] = gen.init[s]("1")
        patterns['init2'] = gen.init[r]("2")
        patterns['fini1'] = gen.fini[s]("1")
        patterns['fini2'] = gen.fini[r]("2")
        patterns['operation1'] = gen.operation[s]("1")
        patterns['operation2'] = gen.operation[r]("2")
        patterns['change_cond'] = '(rank == 1) && (its_raining)'

        # Generate the incorrect matching because of the conditional
        replace = patterns.copy()
        replace['shortdesc'] = 'Point to point @{r}@ is never called.'
        replace['longdesc'] = 'Point to point @{r}@ is never executed. Process 1 calls MPI_Finalize and causes a deadlock.'
        replace['outcome'] = 'ERROR: CallMatching'
        replace['errormsg'] = 'P2P mistmatch. @{r}@ at @{filename}@:@{line:MBIERROR2}@ is never called because of the conditional (@{change_cond}@).'
        replace['operation1'] = gen.operation[s]("1")
        replace['operation2'] = gen.operation[r]("2")
        gen.make_file(template, f'CallOrdering_{r}_{s}_nok.c', replace)



# Copyright (c) 2017-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
This example does not much: It just spans over-polite actor that yield a large amount
of time before ending.

This serves as an example for the simgrid.yield() function, with which an actor can request
to be rescheduled after the other actor that are ready at the current timestamp.

It can also be used to benchmark our context-switching mechanism.
"""

import sys
from simgrid import Actor, Engine, Host, this_actor

def yielder(number_of_yields):
    for _ in range(number_of_yields):
        this_actor.yield_()
    this_actor.info("I yielded {:d} times. Goodbye now!".format(number_of_yields))

if __name__ == '__main__':
    e = Engine(sys.argv)

    e.load_platform(sys.argv[1])             # Load the platform description

    Actor.create("yielder", Host.by_name("Tremblay"), yielder, 10)
    Actor.create("yielder", Host.by_name("Ruby"), yielder, 15)

    e.run()  # - Run the simulation



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

""" This example shows how to suspend and resume an asynchronous communication.
"""

from argparse import ArgumentParser
import sys

from simgrid import Actor, Comm, Engine, Mailbox, this_actor


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        '--platform',
        type=str,
        required=True,
        help='path to the platform description'
    )
    return parser


def sender():
    mailbox: Mailbox = Mailbox.by_name("receiver")
    payload = "Sent message"

    # Create a communication representing the ongoing communication and then
    simulated_size_in_bytes = 13194230
    comm: Comm = mailbox.put_init(payload, simulated_size_in_bytes)
    this_actor.info(f"Suspend the communication before it starts (remaining: {comm.remaining:.0f} bytes)"
                    f" and wait a second.")
    this_actor.sleep_for(1)
    this_actor.info(f"Now, start the communication (remaining: {comm.remaining:.0f} bytes) and wait another second.")
    comm.start()
    this_actor.sleep_for(1)
    this_actor.info(f"There is still {comm.remaining:.0f} bytes to transfer in this communication."
                    " Suspend it for one second.")
    comm.suspend()
    this_actor.info(f"Now there is {comm.remaining:.0f} bytes to transfer. Resume it and wait for its completion.")
    comm.resume()
    comm.wait()
    this_actor.info(f"There is {comm.remaining:.0f} bytes to transfer after the communication completion.")
    this_actor.info(f"Suspending a completed activity is a no-op.")
    comm.suspend()


def receiver():
    mailbox: Mailbox = Mailbox.by_name("receiver")
    this_actor.info("Wait for the message.")
    received: str = mailbox.get()
    this_actor.info(f"I got '{received}'.")


def main():
    settings = create_parser().parse_known_args()[0]
    e = Engine(sys.argv)
    e.load_platform(settings.platform)
    Actor.create("sender", e.host_by_name("Tremblay"), sender)
    Actor.create("receiver", e.host_by_name("Jupiter"), receiver)
    e.run()


if __name__ == "__main__":
    main()



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
This example implements the following scenario:
- Multiple workers consume jobs (Job) from a shared mailbox (worker)
- A client first dispatches several jobs (with a simulated 'cost' - i.e. time the worker will 'process' the job)
- The client then waits for all job results for a maximum time set by the 'wait timeout' (Comm.wait_all_for)
- The client then displays the status of individual jobs.
"""


from argparse import ArgumentParser
from dataclasses import dataclass
from typing import List
from uuid import uuid4
import sys

from simgrid import Actor, Comm, Engine, Host, Mailbox, PyGetAsync, this_actor


SIMULATED_JOB_SIZE_BYTES = 1024
SIMULATED_RESULT_SIZE_BYTES = 1024 * 1024


def parse_requests(requests_str: str) -> List[float]:
    return [float(item.strip()) for item in requests_str.split(",")]


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        '--platform',
        type=str,
        required=True,
        help='path to the platform description'
    )
    parser.add_argument(
        "--workers",
        type=int,
        default=1,
        help="number of worker actors to start"
    )
    parser.add_argument(
        "--jobs",
        type=parse_requests,
        default="1,2,3,4,5",
        help="duration of individual jobs sent to the workers by the client"
    )
    parser.add_argument(
        "--wait-timeout",
        type=float,
        default=5.0,
        help="number of seconds before the client gives up waiting for results and aborts the simulation"
    )
    return parser


@dataclass
class Job:
    job_id: str
    duration: float
    result_mailbox: Mailbox


def worker(worker_id: str):
    this_actor.info(f"{worker_id} started")
    mailbox: Mailbox = Mailbox.by_name("worker")
    while True:
        job: Job = mailbox.get()
        this_actor.info(f"{worker_id} working on {job.job_id} (will take {job.duration}s to complete)")
        this_actor.sleep_for(job.duration)
        job.result_mailbox.put(f"{worker_id}", SIMULATED_RESULT_SIZE_BYTES)


@dataclass
class AsyncJobResult:
    job: Job
    result_comm: Comm
    async_data: PyGetAsync

    @property
    def complete(self) -> bool:
        return self.result_comm.test()

    @property
    def status(self) -> str:
        return "complete" if self.complete else "pending"


def client(client_id: str, jobs: List[float], wait_timeout: float):
    worker_mailbox: Mailbox = Mailbox.by_name("worker")
    this_actor.info(f"{client_id} started")
    async_job_results: list[AsyncJobResult] = []
    for job_idx, job_duration in enumerate(jobs):
        result_mailbox: Mailbox = Mailbox.by_name(str(uuid4()))
        job = Job(job_id=f"job-{job_idx}", duration=job_duration, result_mailbox=result_mailbox)
        out_comm = worker_mailbox.put_init(Job(
            job_id=f"job-{job_idx}",
            duration=job_duration,
            result_mailbox=result_mailbox
        ), SIMULATED_JOB_SIZE_BYTES)
        out_comm.detach()
        result_comm, async_data = result_mailbox.get_async()
        async_job_results.append(AsyncJobResult(
            job=job,
            result_comm=result_comm,
            async_data=async_data
        ))
    this_actor.info(f"awaiting results for all jobs (timeout={wait_timeout}s)")
    completed_comms = Comm.wait_all_for([entry.result_comm for entry in async_job_results], wait_timeout)
    logger = this_actor.warning if completed_comms < len(async_job_results) else this_actor.info
    logger(f"received {completed_comms}/{len(async_job_results)} results")
    for result in async_job_results:
        this_actor.info(f"{result.job.job_id}"
                        f" status={result.status}"
                        f" result_payload={result.async_data.get() if result.complete else ''}")


def main():
    settings = create_parser().parse_known_args()[0]
    e = Engine(sys.argv)
    e.load_platform(settings.platform)
    Actor.create("client", Host.by_name("Tremblay"), client, "client", settings.jobs, settings.wait_timeout)
    for worker_idx in range(settings.workers):
        Actor.create("worker", Host.by_name("Ruby"), worker, f"worker-{worker_idx}").daemonize()
    e.run()


if __name__ == "__main__":
    main()



#! /usr/bin/python3

# Copyright 2021-2022. The MBI project. All rights reserved.
# This program is free software; you can redistribute it and/or modify it under the terms of the license (GNU GPL).

import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: MBI

  Description: @{shortdesc}@
    @{longdesc}@

   Version of MPI: Conforms to MPI 1.1, does not require MPI 2 implementation

BEGIN_MPI_FEATURES
  P2P!basic: Lacking
  P2P!nonblocking: Lacking
  P2P!persistent: Lacking
  COLL!basic: @{collfeature}@
  COLL!nonblocking: @{icollfeature}@
  COLL!persistent: Lacking
  COLL!tools: Lacking
  RMA: Lacking
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 2 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

#define buff_size 128

int main(int argc, char **argv) {
  int nprocs = -1;
  int rank = -1;
  int root = 0;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (nprocs < 2)
    printf("MBI ERROR: This test needs at least 2 processes to produce a bug.\\n");

  MPI_Comm newcom = MPI_COMM_WORLD;
  MPI_Datatype type = MPI_INT;
  MPI_Op op = MPI_SUM;

  int dbs = sizeof(int)*nprocs; /* Size of the dynamic buffers for alltoall and friends */
  @{init1}@
  @{init2}@

  if (@{change_cond}@) {
    @{operation1a}@ /* MBIERROR1 */
    @{fini1a}@
    @{operation2a}@
    @{fini2a}@
  } else {
    @{operation1b}@ /* MBIERROR2 */
    @{fini1b}@
    @{operation2b}@
    @{fini2b}@
  }

  @{free1}@
  @{free2}@

  MPI_Finalize();
  printf("Rank %d finished normally\\n", rank);
  return 0;
}
"""

for c1 in gen.coll + gen.icoll + gen.ibarrier:
    for c2 in gen.coll + gen.icoll + gen.ibarrier:
        patterns = {}
        patterns = {'c1': c1, 'c2': c2}
        patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
        patterns['collfeature'] = 'Yes' if c1 in gen.coll or c2 in gen.coll else 'Lacking'
        patterns['icollfeature'] = 'Yes' if c1 in gen.icoll + gen.ibarrier or c2 in gen.icoll + gen.ibarrier else 'Lacking'
        patterns['c1'] = c1
        patterns['c2'] = c2
        patterns['init1'] = gen.init[c1]("1")
        patterns['init2'] = gen.init[c2]("2")
        patterns['fini1a'] = gen.fini[c1]("1")
        patterns['fini2a'] = gen.fini[c2]("2")
        patterns['fini1b'] = gen.fini[c1]("1")
        patterns['fini2b'] = gen.fini[c2]("2")
        patterns['free1'] = gen.free[c1]("1")
        patterns['free2'] = gen.free[c2]("2")
        patterns['operation1a'] = gen.operation[c1]("1")
        patterns['operation1b'] = gen.operation[c1]("1")
        patterns['operation2a'] = gen.operation[c2]("2")
        patterns['operation2b'] = gen.operation[c2]("2")
        patterns['change_cond'] = 'rank % 2'
        shortdesc = ' collective ordering'

        if c1 == c2:
            # Generate the correct code using the same collective twice
            replace = patterns.copy()
            replace['shortdesc'] = 'Correct' + shortdesc
            replace['longdesc'] = f'All ranks call {c1} twice'
            replace['outcome'] = 'OK'
            replace['errormsg'] = ''
            replace['change_cond'] = 'rank < nprocs'
            replace['operation1b'] = ''
            replace['operation2b'] = ''
            replace['fini1b'] = ''
            replace['fini2b'] = ''
            gen.make_file(template, f'CallOrdering_{c1}_{c2}_ok.c', replace)
            # Generate the correct code using the collective once
            replace = patterns.copy()
            replace['shortdesc'] = 'Correct' + shortdesc
            replace['longdesc'] = f'All ranks call {c1} once'
            replace['outcome'] = 'OK'
            replace['errormsg'] = ''
            replace['init2'] = ''
            replace['change_cond'] = 'rank < nprocs'
            replace['operation2a'] = ''
            replace['operation1b'] = ''
            replace['operation2b'] = ''
            replace['fini2a'] = ''
            replace['fini1b'] = ''
            replace['fini2b'] = ''
            replace['free2'] = ''
            gen.make_file(template, f'CallOrdering_{c1}_ok.c', replace)
        else:
            # Generate the correct ordering with two different collectives
            replace = patterns.copy()
            replace['shortdesc'] = 'Correct' + shortdesc
            replace['longdesc'] = f'All ranks call {c1} and then {c2}'
            replace['outcome'] = 'OK'
            replace['errormsg'] = ''
            replace['change_cond'] = 'rank < nprocs'
            replace['operation1b'] = ''
            replace['operation2b'] = ''
            replace['fini1b'] = ''
            replace['fini2b'] = ''
            gen.make_file(template, f'CallOrdering_{c1}_{c2}_ok.c', replace)
            # Generate the incorrect ordering with two different collectives
            replace = patterns.copy()
            replace['shortdesc'] = 'Incorrect' + shortdesc
            replace['longdesc'] = f'Odd ranks call {c1} and then {c2} while even ranks call these collectives in the other order'
            replace['outcome'] = 'ERROR: CallMatching'
            replace['errormsg'] = 'Collective mistmatch. @{c1}@ at @{filename}@:@{line:MBIERROR1}@ is matched with @{c2}@ line @{filename}@:@{line:MBIERROR2}@.'
            replace['operation1b'] = gen.operation[c2]("2")  # Inversion
            replace['operation2b'] = gen.operation[c1]("1")
            replace['fini1a'] = gen.fini[c1]("1") # Inversion
            replace['fini2a'] = gen.fini[c2]("2")
            replace['fini1b'] = gen.fini[c2]("2") # Inversion
            replace['fini2b'] = gen.fini[c1]("1")
            replace['free1'] = gen.free[c2]("2")
            replace['free2'] = gen.free[c1]("1")

            gen.make_file(template, f'CallOrdering_{c1}_{c2}_nok.c', replace)

    # Generate the incorrect ordering with one collective
    replace = patterns.copy()
    replace['shortdesc'] = 'Incorrect' + shortdesc
    replace['longdesc'] = f'Odd ranks call {c1} while even ranks do not call any collective'
    replace['outcome'] = 'ERROR: CallMatching'
    replace['errormsg'] = 'Collective mistmatch. @{c1}@ at @{filename}@:@{line:MBIERROR1}@ is not matched.'
    replace['operation1b'] = ''  # Remove functions
    replace['operation2b'] = ''
    replace['operation2a'] = ''
    replace['init2'] = ''
    replace['fini1b'] = ''
    replace['fini2a'] = ''
    replace['fini2b'] = ''
    replace['free1'] = gen.free[c1]("1")
    replace['free2'] = ''
    gen.make_file(template, f'CallOrdering_{c1}_none_nok.c', replace)
    # Generate a correct ordering with a conditional not depending on ranks
    replace = patterns.copy()
    replace['shortdesc'] = 'Correct' + shortdesc
    replace['longdesc'] = f'All ranks call {c1}'
    replace['outcome'] = 'OK'
    replace['errormsg'] = ''
    replace['change_cond'] = 'rank < nprocs'
    replace['operation1b'] = '' # Remove functions
    replace['operation2b'] = ''
    replace['operation2a'] = ''
    replace['init2'] = ''
    replace['fini1b'] = ''
    replace['fini2a'] = ''
    replace['fini2b'] = ''
    replace['free1'] = gen.free[c1]("1")
    replace['free2'] = ''
    gen.make_file(template, f'CallOrdering_{c1}_none_ok.c', replace)



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

""" This example shows how to suspend and resume an asynchronous communication.
"""

from argparse import ArgumentParser
from typing import List
import sys

from simgrid import Actor, Comm, Engine, Mailbox, this_actor


FINALIZE_MESSAGE = "finalize"


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        '--platform',
        type=str,
        required=True,
        help='path to the platform description'
    )
    return parser


def sender(receiver_mailbox: Mailbox, messages_count: int, payload_size: int):
    pending_comms: List[Comm] = []
    # Start dispatching all messages to the receiver
    for i in range(messages_count):
        payload = f"Message {i}"
        this_actor.info(f"Send '{payload}' to '{receiver_mailbox.name}'")
        # Create a communication representing the ongoing communication
        comm = receiver_mailbox.put_async(payload, payload_size)
        # Add this comm to the vector of all known comms
        pending_comms.append(comm)

    # Start the finalize signal to the receiver
    final_comm = receiver_mailbox.put_async(FINALIZE_MESSAGE, 0)
    pending_comms.append(final_comm)
    this_actor.info(f"Send '{FINALIZE_MESSAGE}' to '{receiver_mailbox.name}'")
    this_actor.info("Done dispatching all messages")

    # Now that all message exchanges were initiated, wait for their completion, in order of creation
    while pending_comms:
        comm = pending_comms[-1]
        comm.wait_until(Engine.clock + 1)
        pending_comms.pop()  # remove it from the list
    this_actor.info("Goodbye now!")


def receiver(mailbox: Mailbox):
    this_actor.info("Wait for my first message")
    finalized = False
    while not finalized:
        received: str = mailbox.get()
        this_actor.info(f"I got a '{received}'.")
        # If it's a finalize message, we're done.
        if received == FINALIZE_MESSAGE:
            finalized = True


def main():
    settings = create_parser().parse_known_args()[0]
    e = Engine(sys.argv)
    e.load_platform(settings.platform)
    receiver_mailbox: Mailbox = Mailbox.by_name("receiver")
    Actor.create("sender", e.host_by_name("Tremblay"), sender, receiver_mailbox, 3, int(5e7))
    Actor.create("receiver", e.host_by_name("Ruby"), receiver, receiver_mailbox)
    e.run()


if __name__ == "__main__":
    main()



#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: MBI

  Description: @{shortdesc}@
    @{longdesc}@

   Version of MPI: Conforms to MPI 1.1, does not require MPI 2 implementation

BEGIN_MPI_FEATURES
  P2P!basic: Lacking
  P2P!nonblocking: Lacking
  P2P!persistent: Lacking
  COLL!basic: Lacking
  COLL!nonblocking: Lacking
  COLL!persistent: Lacking
  COLL!tools: @{toolfeature}@
  RMA: Lacking
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 2 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

#define buff_size 128

int main(int argc, char **argv) {
  int nprocs = -1;
  int rank = -1;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (nprocs < 2)
    printf("MBI ERROR: This test needs at least 2 processes to produce a bug!\\n");

   MPI_Comm newcom;
   int dims[2], periods[2], coords[2];
   int source, dest;
   dims[0] = 2;
   dims[1] = 1;
   periods[0] = 1;
   periods[1] = 1;

   @{change_dims}@

   MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, 0, &newcom); /* create a cartesian communicator */

   @{change_com}@

   @{init}@
   @{operation}@ /* MBIERROR2 */
   @{fini}@

   if (newcom != MPI_COMM_NULL)
     MPI_Comm_free(&newcom);

  MPI_Finalize();
  printf("Rank %d finished normally\\n", rank);
  return 0;
}
"""

for c in gen.tcoll4topo:
    patterns = {}
    patterns = {'c': c}
    patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
    patterns['toolfeature'] = 'Yes' if c in gen.tcoll4topo else 'Lacking'
    patterns['c'] = c
    patterns['init'] = gen.init[c]("1")
    patterns['fini'] = gen.fini[c]("1")
    patterns['operation'] = gen.operation[c]("1")
    patterns['change_dims'] = '/* No error injected here */'

    # Generate the correct code
    replace = patterns.copy()
    replace['shortdesc'] = 'Function @{c}@ with correct arguments'
    replace['longdesc'] = f'All ranks in comm call {c} with correct arguments'
    replace['outcome'] = 'OK'
    replace['errormsg'] = ''
    replace['change_com'] = '/* No error injected here */'
    replace['change_dims'] = '/* No error injected here */'
    gen.make_file(template, f'InvalidParam_{c}_ok.c', replace)

    # Generate the incorrect code
    replace = patterns.copy()
    replace['shortdesc'] = 'The code tries to get cartesian information of MPI_COMM_WORLD.'
    replace['longdesc'] = 'The code creates a cartesian communicator, and tries to get cartesian information of MPI_COMM_WORLD.'
    replace['outcome'] = 'ERROR: InvalidCommunicator'
    replace['errormsg'] = 'Invalid Communicator in a collective. @{c}@ at @{filename}@:@{line:MBIERROR2}@ tries to get cartesian information of MPI_COMM_WORLD.'
    replace['change_com'] = 'newcom = MPI_COMM_WORLD; /* MBIERROR1 */'
    gen.make_file(template, f'InvalidParam_Com_{c}_nok.c', replace)

    # Generate the code with newcom=MPI_COMM_NULL
    replace = patterns.copy()
    replace['shortdesc'] = 'Function @{c}@ called with comm=MPI_COMM_NULL'
    replace['longdesc'] = 'Function @{c}@ called with comm=MPI_COMM_NULL'
    replace['outcome'] = 'ERROR: InvalidCommunicator'
    replace['errormsg'] = 'Invalid communicator. @{c}@ at @{filename}@:@{line:MBIERROR2}@ has MPI_COMM_NULL as a communicator.'
    replace['change_com'] = 'newcom = MPI_COMM_NULL; /* MBIERROR1 */'
    gen.make_file(template, f'InvalidParam_ComNull_{c}_nok.c', replace)

    # Generate the code with invalid dimension
    replace = patterns.copy()
    replace['shortdesc'] = 'Creates a cartesian communicator with a negative entry in the dims attribute'
    replace['longdesc'] = 'Creates a cartesian communicator with a negative entry in the dims attribute, which is a usage error'
    replace['outcome'] = 'ERROR: InvalidOtherArg'
    replace['errormsg'] = 'Invalid Argument. MPI_Cart_create has invalid dimensions.'
    replace['change_com'] = ""
    replace['change_dims'] = 'dims[0] = -2; dims[1] = -1; /* MBIERROR1 */'
    gen.make_file(template, 'InvalidParam_Dim_MPI_Cart_create_nok.c', replace)



# Copyright (c) 2019-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
Second failing example for bug #9 on Framagit (Python bindings crashing)

An intricate recursion is used to make the crash happen.
"""

import sys
from simgrid import Engine, this_actor

def do_sleep1(i, dur):
    if i > 0:
        this_actor.info("1-Iter {:d}".format(i))
        do_sleep3(i - 1, dur)
        this_actor.sleep_for(dur)
        this_actor.info("1-Mid ({:d})".format(i))
        do_sleep3(int(i / 2), dur)
        this_actor.info("1-Done ({:d})".format(i))

def do_sleep3(i, dur):
    if i > 0:
        this_actor.info("3-Iter {:d}".format(i))
        do_sleep5(i - 1, dur)
        this_actor.sleep_for(dur)
        this_actor.info("3-Mid ({:d})".format(i))
        do_sleep5(int(i / 2), dur)
        this_actor.info("3-Done ({:d})".format(i))

def do_sleep5(i, dur):
    if i > 0:
        this_actor.info("5-Iter {:d}".format(i))
        do_sleep1(i - 1, dur)
        this_actor.sleep_for(dur)
        this_actor.info("5-Mid ({:d})".format(i))
        do_sleep1(int(i / 2), dur)
        this_actor.info("5-Done ({:d})".format(i))

def sleeper1():
    do_sleep1(16, 1)

def sleeper3():
    do_sleep3(6, 3)

def sleeper5():
    do_sleep5(4, 5)

if __name__ == '__main__':
    e = Engine(sys.argv)

    e.load_platform(sys.argv[1])             # Load the platform description

    # Register the classes representing the actors
    e.register_actor("sleeper1", sleeper1)
    e.register_actor("sleeper3", sleeper3)
    e.register_actor("sleeper5", sleeper5)

    e.load_deployment(sys.argv[2])

    e.run()
    this_actor.info("Finalize!")



# Copyright (c) 2006-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

import sys
from simgrid import Actor, Engine, Host, Mailbox, NetZone, LinkInRoute, this_actor

class Sender:
    """
    Send 1 message for each host
    """
    def __init__(self, hosts):
        self.hosts = hosts

    # Actors that are created as object will execute their __call__ method.
    # So, the following constitutes the main function of the Sender actor.
    def __call__(self):

        for host in self.hosts:
            mbox = Mailbox.by_name(host.name)
            msg = "Hello. I'm " + str(this_actor.get_host().name)
            size = int(1e6)
            this_actor.info("Sending msg to " + host.name)
            mbox.put(msg, size)

        this_actor.info("Done dispatching all messages. Goodbye!")

class Receiver:
    """
    Receiver actor: wait for 1 messages and do operations
    """

    def __call__(self):
        this_actor.execute(1e9)
        for disk in Host.current().get_disks():
            this_actor.info("Using disk " + disk.name)
            disk.read(10000)
            disk.write(10000)
        mbox = Mailbox.by_name(this_actor.get_host().name)
        msg = mbox.get()
        this_actor.info("I got '%s'." % msg)
        this_actor.info("Finished executing. Goodbye!")

def load_platform():
    """ Creates a mixed platform, using many methods available in the API
    """

    root = NetZone.create_floyd_zone("root")
    hosts = []
    # dijkstra
    dijkstra = NetZone.create_dijkstra_zone("dijkstra")
    msg_base = "Creating zone: "
    this_actor.info(msg_base + dijkstra.name)
    dijkstra.set_parent(root)
    host1 = dijkstra.create_host("host1", [1e9, 1e8])
    host1.core_count = 2
    hosts.append(host1)
    host1.create_disk("disk1", 1e5, 1e4).seal()
    host1.create_disk("disk2", "1MBps", "1Mbps").seal()
    host1.seal()
    host2 = dijkstra.create_host("host2", ["1Gf", "1Mf"]).seal()
    hosts.append(host2)
    link1 = dijkstra.create_link("link1_up", [1e9]).set_latency(1e-3).set_concurrency_limit(10).seal()
    link2 = dijkstra.create_link("link1_down", ["1GBps"]).set_latency("1ms").seal()
    dijkstra.add_route(host1.netpoint, host2.netpoint, None, None, [LinkInRoute(link1)], False)
    dijkstra.add_route(host2.netpoint, host1.netpoint, None, None, [LinkInRoute(link2)], False)
    dijkstra.seal()

    # vivaldi
    vivaldi = NetZone.create_vivaldi_zone("vivaldi")
    this_actor.info(msg_base + vivaldi.name)
    vivaldi.set_parent(root)
    host3 = vivaldi.create_host("host3", 1e9).set_coordinates("1 1 1").seal()
    host4 = vivaldi.create_host("host4", "1Gf").set_coordinates("2 2 2").seal()
    hosts.append(host3)
    hosts.append(host4)

    # empty
    empty = NetZone.create_empty_zone("empty")
    this_actor.info(msg_base + empty.name)
    empty.set_parent(root)
    host5 = empty.create_host("host5", 1e9)
    hosts.append(host5)
    empty.seal()

    # wifi
    wifi = NetZone.create_wifi_zone("wifi")
    this_actor.info(msg_base + wifi.name)
    wifi.set_parent(root)
    router = wifi.create_router("wifi_router")
    wifi.set_property("access_point", "wifi_router")
    host6 = wifi.create_host(
        "host6", ["100.0Mf", "50.0Mf", "20.0Mf"]).seal()
    hosts.append(host6)
    wifi_link = wifi.create_link("AP1", ["54Mbps", "36Mbps", "24Mbps"]).seal()
    wifi_link.set_host_wifi_rate(host6, 1)
    wifi.seal()

    # create routes between netzones
    link_a = vivaldi.create_link("linkA", 1e9).seal()
    link_b = vivaldi.create_link("linkB", "1GBps").seal()
    link_c = vivaldi.create_link("linkC", "1GBps").seal()
    root.add_route(dijkstra.netpoint, vivaldi.netpoint, host1.netpoint, host3.netpoint, [LinkInRoute(link_a)], True)
    root.add_route(vivaldi.netpoint, empty.netpoint, host3.netpoint, host5.netpoint, [LinkInRoute(link_b)], True)
    root.add_route(empty.netpoint, wifi.netpoint, host5.netpoint, router, [LinkInRoute(link_c)], True)

    # create actors Sender/Receiver
    Actor.create("sender", hosts[0], Sender(hosts))
    for host in hosts:
        Actor.create("receiver", host, Receiver())

###################################################################################################

if __name__ == '__main__':
    e = Engine(sys.argv)

    # create platform
    load_platform()

    # runs the simulation
    e.run()



# Copyright 2021-2022. The MBI project. All rights reserved.
# This program is free software; you can redistribute it and/or modify it under the terms of the license (GNU GPL).

# This is a simple templating system, dedicated to the systematic generation of MPI source code

import os
import re

# Collectives
coll = ['MPI_Barrier', 'MPI_Bcast', 'MPI_Reduce', 'MPI_Gather', 'MPI_Scatter', 'MPI_Scan', 'MPI_Exscan', 'MPI_Allgather', 'MPI_Allreduce', 'MPI_Allgatherv', 'MPI_Alltoall', 'MPI_Alltoallv']
icoll = ['MPI_Ibcast', 'MPI_Ireduce', 'MPI_Igather', 'MPI_Iscatter', 'MPI_Iscan', 'MPI_Iexscan', 'MPI_Iallgather', 'MPI_Iallreduce', 'MPI_Iallgatherv', 'MPI_Ialltoall', 'MPI_Ialltoallv']
ibarrier = ['MPI_Ibarrier']
coll4op = ['MPI_Reduce', 'MPI_Allreduce']
icoll4op = ['MPI_Ireduce', 'MPI_Iallreduce']
coll4root = ['MPI_Reduce', 'MPI_Bcast', 'MPI_Gather', 'MPI_Scatter']
icoll4root = ['MPI_Ireduce', 'MPI_Ibcast', 'MPI_Igather', 'MPI_Iscatter']
pcoll = []
tcoll = ['MPI_Comm_split', 'MPI_Op_create', 'MPI_Comm_dup', 'MPI_Type_contiguous', 'MPI_Comm_create', 'MPI_Group_excl'] # MPI_Comm_dup removed
tcoll4color = ['MPI_Comm_split']
tcoll4topo = ['MPI_Cart_get']

# P2P
allsend = ['MPI_Send', 'MPI_Isend', 'MPI_Ssend', 'MPI_Bsend', 'MPI_Send_init']
allrecv = ['MPI_Recv', 'MPI_Irecv', 'MPI_Recv_init']
send = ['MPI_Send']
ssend = ['MPI_Ssend']
bsend = ['MPI_Bsend']
isend = ['MPI_Isend']
psend = ['MPI_Send_init']
recv = ['MPI_Recv']
irecv = ['MPI_Irecv']
precv = ['MPI_Recv_init']
probe = ['MPI_Probe']

# RMA
rma = ['MPI_Get', 'MPI_Put']
get = ['MPI_Get']
put = ['MPI_Put']
store = ['store']
load = ['load']
rstore = ['rstore']
rload = ['rload']
loadstore = ['loadstore']


# setup
init = {}
start = {}
operation = {}
fini = {}
free = {}
write = {}
error = {}
epoch = {}
finEpoch = {}


### COLL:basic

init['MPI_Bcast'] = lambda n: f'int buf{n}[buff_size];'
start['MPI_Bcast'] = lambda n: ""
operation['MPI_Bcast'] = lambda n: f'MPI_Bcast(buf{n}, buff_size, type, root, newcom);'
fini['MPI_Bcast'] = lambda n: ""
free['MPI_Bcast'] = lambda n: ""
write['MPI_Bcast'] = lambda n: ""

init['MPI_Barrier'] = lambda n: ""
start['MPI_Barrier'] = lambda n: ""
operation['MPI_Barrier'] = lambda n: 'MPI_Barrier(newcom);'
fini['MPI_Barrier'] = lambda n: ""
free['MPI_Barrier'] = lambda n: ""
write['MPI_Barrier'] = lambda n: ""

init['MPI_Reduce'] = lambda n: f"int sum{n}, val{n} = 1;"
start['MPI_Reduce'] = lambda n: ""
operation['MPI_Reduce'] = lambda n: f"MPI_Reduce(&val{n}, &sum{n}, 1, type, op, root, newcom);"
fini['MPI_Reduce'] = lambda n: ""
free['MPI_Reduce'] = lambda n: ""
write['MPI_Reduce'] = lambda n: ""

init['MPI_Gather'] = lambda n: f"int val{n}=1, buf{n}[buff_size];"
start['MPI_Gather'] = lambda n: ""
operation['MPI_Gather'] = lambda n: f"MPI_Gather(&val{n}, 1, type, buf{n},1, type, root, newcom);"
fini['MPI_Gather'] = lambda n: ""
free['MPI_Gather'] = lambda n: ""
write['MPI_Gather'] = lambda n: ""

init['MPI_Scatter'] = lambda n: f"int val{n}, buf{n}[buff_size];\n  memset(buf{n}, 0, sizeof(int)*buff_size);"
start['MPI_Scatter'] = lambda n: ""
operation['MPI_Scatter'] = lambda n: f"MPI_Scatter(&buf{n}, 1, type, &val{n}, 1, type, root, newcom);"
fini['MPI_Scatter'] = lambda n: ""
free['MPI_Scatter'] = lambda n: ""
write['MPI_Scatter'] = lambda n: ""

init['MPI_Allreduce'] = lambda n: f"int sum{n}, val{n} = 1;"
start['MPI_Allreduce'] = lambda n: ""
operation['MPI_Allreduce'] = lambda n: f"MPI_Allreduce(&val{n}, &sum{n}, 1, type, op, newcom);"
fini['MPI_Allreduce'] = lambda n: ""
free['MPI_Allreduce'] = lambda n: ""
write['MPI_Allreduce'] = lambda n: ""

init['MPI_Scan'] = lambda n: f"int outbuf{n}[buff_size];\n  memset(outbuf{n}, 0, buff_size*sizeof(int));\n  int inbuf{n}[buff_size];"
start['MPI_Scan'] = lambda n: ""
operation['MPI_Scan'] = lambda n: f"MPI_Scan(&outbuf{n}, inbuf{n}, buff_size, type, op, newcom);"
fini['MPI_Scan'] = lambda n: ""
free['MPI_Scan'] = lambda n: ""
write['MPI_Scan'] = lambda n: ""

init['MPI_Exscan'] = lambda n: f"int outbuf{n}[buff_size];\n  memset(outbuf{n}, 0, buff_size*sizeof(int));\n  int inbuf{n}[buff_size];"
start['MPI_Exscan'] = lambda n: ""
operation['MPI_Exscan'] = lambda n: f"MPI_Exscan(&outbuf{n}, inbuf{n}, buff_size, type, op, newcom);"
fini['MPI_Exscan'] = lambda n: ""
free['MPI_Exscan'] = lambda n: ""
write['MPI_Exscan'] = lambda n: ""

init['MPI_Allgather'] = lambda n: f"int val{n}=1, *rbuf{n} = (int*)malloc(dbs);"
start['MPI_Allgather'] = lambda n: ""
operation['MPI_Allgather'] = lambda n: f"MPI_Allgather(&val{n}, 1, type, rbuf{n}, 1, type, newcom);"
fini['MPI_Allgather'] = lambda n: ""
free['MPI_Allgather'] = lambda n: f"free(rbuf{n});"
write['MPI_Allgather'] = lambda n: ""

init['MPI_Alltoallv'] = lambda n: (f"int *sbuf{n}=(int*)malloc(dbs*2), *rbuf{n}=(int*)malloc(dbs*2), *scounts{n}=(int*)malloc(dbs), *rcounts{n}=(int*)malloc(dbs), *sdispls{n}=(int*)malloc(dbs), *rdispls{n}=(int*)malloc(dbs);\n"
  +  "  for (int i = 0; i < nprocs; i++) {\n"
  + f"    scounts{n}[i] = 2;\n"
  + f"    rcounts{n}[i] = 2;\n"
  + f"    sdispls{n}[i] = (nprocs - (i + 1)) * 2;\n"
  + f"    rdispls{n}[i] = i * 2;\n"
  +  "  }")
start['MPI_Alltoallv'] = lambda n: ""
operation['MPI_Alltoallv'] = lambda n: f"MPI_Alltoallv(sbuf{n}, scounts{n}, sdispls{n}, type, rbuf{n}, rcounts{n}, rdispls{n}, type, newcom);"
fini['MPI_Alltoallv'] = lambda n: ""
free['MPI_Alltoallv'] = lambda n: f"free(sbuf{n});free(rbuf{n});free(scounts{n});free(rcounts{n});free(sdispls{n});free(rdispls{n});"
write['MPI_Alltoallv'] = lambda n: ""

init['MPI_Alltoall'] = lambda n: f"int *sbuf{n} = (int*)malloc(dbs), *rbuf{n} = (int*)malloc(dbs);"
start['MPI_Alltoall'] = lambda n: ""
operation['MPI_Alltoall'] = lambda n: f"MPI_Alltoall(sbuf{n}, 1, type, rbuf{n}, 1, type, newcom);"
fini['MPI_Alltoall'] = lambda n: ""
free['MPI_Alltoall'] = lambda n: f"free(sbuf{n});free(rbuf{n});"
write['MPI_Alltoall'] = lambda n: ""

init['MPI_Allgatherv'] = lambda n: (f"int *rbuf{n} = (int*)malloc(dbs*2), *rcounts{n}=(int*)malloc(dbs),  *displs{n}=(int*)malloc(dbs);\n"
  +  "  for (int i = 0; i < nprocs; i++) {\n"
  + f"    rcounts{n}[i] = 1;\n"
  + f"    displs{n}[i] = 2 * (nprocs - (i + 1));\n"
  +  "  }")
start['MPI_Allgatherv'] = lambda n: ""
operation['MPI_Allgatherv'] = lambda n: f"MPI_Allgatherv(&rank, 1, type, rbuf{n}, rcounts{n}, displs{n}, type, newcom);"
fini['MPI_Allgatherv'] = lambda n: ""
free['MPI_Allgatherv'] = lambda n: f"free(rbuf{n});free(rcounts{n});free(displs{n});"
write['MPI_Allgatherv'] = lambda n: ""


### COLL:nonblocking

init['MPI_Ibarrier'] = lambda n: f"MPI_Request req{n}=MPI_REQUEST_NULL; MPI_Status stat{n};"
start['MPI_Ibarrier'] = lambda n: ""
operation['MPI_Ibarrier'] = lambda n: f'MPI_Ibarrier(newcom, &req{n});'
fini['MPI_Ibarrier'] = lambda n: f"MPI_Wait(&req{n}, &stat{n});"
free['MPI_Ibarrier'] = lambda n: f'if(req{n} != MPI_REQUEST_NULL) MPI_Request_free(&req{n});'
write['MPI_Ibarrier'] = lambda n: ""

init['MPI_Ireduce'] = lambda n: f"MPI_Request req{n}=MPI_REQUEST_NULL; MPI_Status stat{n}; int sum{n}, val{n} = 1;"
start['MPI_Ireduce'] = lambda n: ""
operation['MPI_Ireduce'] = lambda n: f"MPI_Ireduce(&val{n}, &sum{n}, 1, type, op, root, newcom, &req{n});"
fini['MPI_Ireduce'] = lambda n: f"MPI_Wait(&req{n}, &stat{n});"
free['MPI_Ireduce'] = lambda n: f'if(req{n} != MPI_REQUEST_NULL) MPI_Request_free(&req{n});'
write['MPI_Ireduce'] = lambda n: f"sum{n}++;"

init['MPI_Iallreduce'] = lambda n: f'MPI_Request req{n}=MPI_REQUEST_NULL;MPI_Status stat{n}; int sum{n}, val{n} = 1;'
start['MPI_Iallreduce'] = lambda n: ""
operation['MPI_Iallreduce'] = lambda n: f'MPI_Iallreduce(&val{n}, &sum{n}, 1, type, op, newcom, &req{n});'
fini['MPI_Iallreduce'] = lambda n: f'MPI_Wait(&req{n}, &stat{n});'
free['MPI_Iallreduce'] = lambda n: f"if(req{n} != MPI_REQUEST_NULL) MPI_Request_free(&req{n});"
write['MPI_Iallreduce'] = lambda n: f"sum{n}++;"

init['MPI_Ibcast'] = lambda n: f'MPI_Request req{n}=MPI_REQUEST_NULL; MPI_Status sta{n};int buf{n}[buff_size];'
start['MPI_Ibcast'] = lambda n: ""
operation['MPI_Ibcast'] = lambda n: f'MPI_Ibcast(buf{n}, buff_size, type, root, newcom, &req{n});'
fini['MPI_Ibcast'] = lambda n: f"MPI_Wait(&req{n},&sta{n});"
free['MPI_Ibcast'] = lambda n: f'if(req{n} != MPI_REQUEST_NULL) MPI_Request_free(&req{n});'
write['MPI_Ibcast'] = lambda n: f'buf{n}[0]++;'

init['MPI_Igather'] = lambda n: f"int val{n}=1, buf{n}[buff_size];MPI_Request req{n}=MPI_REQUEST_NULL;MPI_Status sta{n};"
start['MPI_Igather'] = lambda n: ""
operation['MPI_Igather'] = lambda n: f'MPI_Igather(&val{n}, 1, type, &buf{n},1, type, root, newcom, &req{n});'
write['MPI_Igather'] = lambda n: f'val{n}=3;'
fini['MPI_Igather'] = lambda n: f'MPI_Wait(&req{n},&sta{n});'
free['MPI_Igather'] = lambda n: f'if(req{n} != MPI_REQUEST_NULL) MPI_Request_free(&req{n});'

init['MPI_Iscatter'] = lambda n: f"MPI_Request req{n} = MPI_REQUEST_NULL;\n  MPI_Status sta{n};\n  int val{n};\n  int buf{n}[buff_size];\n  memset(buf{n}, 0, buff_size*sizeof(int));"
start['MPI_Iscatter'] = lambda n: ""
operation['MPI_Iscatter'] = lambda n: f"MPI_Iscatter(&buf{n}, 1, type, &val{n}, 1, type, root, newcom,&req{n});"
fini['MPI_Iscatter'] = lambda n: f"MPI_Wait(&req{n},&sta{n});"
free['MPI_Iscatter'] = lambda n: f'if(req{n} != MPI_REQUEST_NULL) MPI_Request_free(&req{n});'
write['MPI_Iscatter'] = lambda n: f'buf{n}[0]++;'

init['MPI_Iscan'] = lambda n: f"MPI_Request req{n} = MPI_REQUEST_NULL;\n  MPI_Status sta{n};\n  int outbuf{n}[buff_size];\n  memset(outbuf{n}, 0, buff_size*sizeof(int));\n  int inbuf{n}[buff_size];"
start['MPI_Iscan'] = lambda n: ""
operation['MPI_Iscan'] = lambda n: f"MPI_Iscan(&outbuf{n}, inbuf{n}, buff_size, type, op, newcom,&req{n});"
fini['MPI_Iscan'] = lambda n: f"MPI_Wait(&req{n},&sta{n});"
free['MPI_Iscan'] = lambda n: f'if(req{n} != MPI_REQUEST_NULL) MPI_Request_free(&req{n});'
write['MPI_Iscan'] = lambda n: f'outbuf{n}[0]++;'

init['MPI_Iexscan'] = lambda n: f"MPI_Request req{n}=MPI_REQUEST_NULL;MPI_Status sta{n};\n  int outbuf{n}[buff_size];\n  memset(outbuf{n}, 0, buff_size*sizeof(int));\n  int inbuf{n}[buff_size];"
start['MPI_Iexscan'] = lambda n: ""
operation['MPI_Iexscan'] = lambda n: f"MPI_Iexscan(&outbuf{n}, inbuf{n}, buff_size, type, op, newcom,&req{n});"
fini['MPI_Iexscan'] = lambda n: f"MPI_Wait(&req{n},&sta{n});"
free['MPI_Iexscan'] = lambda n: f'if(req{n} != MPI_REQUEST_NULL) MPI_Request_free(&req{n});'
write['MPI_Iexscan'] = lambda n: f'outbuf{n}[0]++;'

init['MPI_Iallgather'] = lambda n: f"MPI_Request req{n}=MPI_REQUEST_NULL;MPI_Status sta{n};int val{n}=1, *rbuf{n} = (int*)malloc(dbs);"
start['MPI_Iallgather'] = lambda n: ""
operation['MPI_Iallgather'] = lambda n: f"MPI_Iallgather(&val{n}, 1, type, rbuf{n}, 1, type, newcom,&req{n});"
fini['MPI_Iallgather'] = lambda n: f"MPI_Wait(&req{n},&sta{n});"
free['MPI_Iallgather'] = lambda n: f"free(rbuf{n});"
write['MPI_Iallgather'] = lambda n: f'val{n}++;'

init['MPI_Iallgatherv'] = lambda n: (f"MPI_Request req{n}=MPI_REQUEST_NULL;MPI_Status sta{n};int *rbuf{n} = (int*)malloc(dbs*2), *rcounts{n}=(int*)malloc(dbs),  *displs{n}=(int*)malloc(dbs);\n"
  +  "  for (int i = 0; i < nprocs; i++) {\n"
  + f"    rcounts{n}[i] = 1;\n"
  + f"    displs{n}[i] = 2 * (nprocs - (i + 1));\n"
  +  "  }")
start['MPI_Iallgatherv'] = lambda n: ""
operation['MPI_Iallgatherv'] = lambda n: f"MPI_Iallgatherv(&rank, 1, type, rbuf{n}, rcounts{n}, displs{n}, type, newcom,&req{n});"
fini['MPI_Iallgatherv'] = lambda n: f"MPI_Wait(&req{n},&sta{n});"
free['MPI_Iallgatherv'] = lambda n: f"free(rbuf{n});free(rcounts{n});free(displs{n});"
write['MPI_Iallgatherv'] = lambda n: f"rbuf{n}[0]++;"

init['MPI_Ialltoall'] = lambda n: f"MPI_Request req{n}=MPI_REQUEST_NULL;MPI_Status sta{n};int *sbuf{n} = (int*)malloc(dbs), *rbuf{n} = (int*)malloc(dbs);"
start['MPI_Ialltoall'] = lambda n: ""
operation['MPI_Ialltoall'] = lambda n: f"MPI_Ialltoall(sbuf{n}, 1, type, rbuf{n}, 1, type, newcom, &req{n});"
fini['MPI_Ialltoall'] = lambda n: f"MPI_Wait(&req{n},&sta{n});"
free['MPI_Ialltoall'] = lambda n: f"free(sbuf{n});free(rbuf{n});"
write['MPI_Ialltoall'] = lambda n: f"rbuf{n}[0]++;"

init['MPI_Ialltoallv'] = lambda n: (f"MPI_Request req{n}=MPI_REQUEST_NULL;MPI_Status sta{n};int *sbuf{n}=(int*)malloc(dbs*2), *rbuf{n}=(int*)malloc(dbs*2), *scounts{n}=(int*)malloc(dbs), *rcounts{n}=(int*)malloc(dbs), *sdispls{n}=(int*)malloc(dbs), *rdispls{n}=(int*)malloc(dbs);\n"
  +  "  for (int i = 0; i < nprocs; i++) {\n"
  + f"    scounts{n}[i] = 2;\n"
  + f"    rcounts{n}[i] = 2;\n"
  + f"    sdispls{n}[i] = (nprocs - (i + 1)) * 2;\n"
  + f"    rdispls{n}[i] = i * 2;\n"
  +  "  }")
start['MPI_Ialltoallv'] = lambda n: ""
operation['MPI_Ialltoallv'] = lambda n: f"MPI_Ialltoallv(sbuf{n}, scounts{n}, sdispls{n}, type, rbuf{n}, rcounts{n}, rdispls{n}, type, newcom,&req{n});"
fini['MPI_Ialltoallv'] = lambda n: f"MPI_Wait(&req{n},&sta{n});"
free['MPI_Ialltoallv'] = lambda n: f"free(sbuf{n});free(rbuf{n});free(scounts{n});free(rcounts{n});free(sdispls{n});free(rdispls{n});"
write['MPI_Ialltoallv'] = lambda n: f"rbuf{n}[0]++;"

### COLL:persistent



### COLL:tools

init['MPI_Comm_split'] = lambda n: 'MPI_Comm com[size]; int color = rank % 2; int key = 1;'
start['MPI_Comm_split'] = lambda n: ""
operation['MPI_Comm_split'] = lambda n: 'MPI_Comm_split(MPI_COMM_WORLD,color,key, &com[j]);'
error['MPI_Comm_split'] = 'CommunicatorLeak'
fini['MPI_Comm_split'] = lambda n: "if(com[j] != MPI_COMM_NULL) MPI_Comm_free(&com[j]);"
free['MPI_Comm_split'] = lambda n: ""


init['MPI_Cart_get'] = lambda n: ""
start['MPI_Cart_get'] = lambda n: ""
operation['MPI_Cart_get'] = lambda n: 'MPI_Cart_get(newcom, 2, dims, periods, coords);'
write['MPI_Cart_get'] = lambda n: ""
fini['MPI_Cart_get'] = lambda n: ""
free['MPI_Cart_get'] = lambda n: ""


init['MPI_Op_create'] = lambda n: 'MPI_Op op[size];'
operation['MPI_Op_create'] = lambda n: 'MPI_Op_create((MPI_User_function *)myOp, 0, &op[j]);'
error['MPI_Op_create'] = 'OperatorLeak'
fini['MPI_Op_create'] = lambda n: "MPI_Op_free(&op[j]);"
free['MPI_Op_create'] = lambda n: ""

init['MPI_Comm_group'] = lambda n: 'MPI_Group grp[size];'
operation['MPI_Comm_group'] = lambda n: 'MPI_Comm_group(MPI_COMM_WORLD, &grp[j]);'
error['MPI_Comm_group'] = 'GroupLeak'
fini['MPI_Comm_group'] = lambda n: "MPI_Group_free(&grp[j]);"
free['MPI_Comm_group'] = lambda n: ""

init['MPI_Group_excl'] = lambda n: 'MPI_Group worldgroup, grp[size];\n MPI_Comm_group(MPI_COMM_WORLD, &worldgroup);'
operation['MPI_Group_excl'] = lambda n: 'MPI_Group_excl(worldgroup, 1, &rank, &grp[j]);'
error['MPI_Group_excl'] = 'GroupLeak'
fini['MPI_Group_excl'] = lambda n: "MPI_Group_free(&grp[j]);"
free['MPI_Group_excl'] = lambda n: "MPI_Group_free(&worldgroup);"

init['MPI_Comm_create'] = lambda n: 'MPI_Comm com[size]; MPI_Group grp[size];'
operation['MPI_Comm_create'] = lambda n: 'MPI_Comm_group(MPI_COMM_WORLD, &grp[j]);\n MPI_Comm_create(MPI_COMM_WORLD, grp[j], &com[j]);\n MPI_Group_free(&grp[j]);'
error['MPI_Comm_create'] = 'CommunicatorLeak'
fini['MPI_Comm_create'] = lambda n: "MPI_Comm_free(&com[j]);"
free['MPI_Comm_create'] = lambda n: ""

init['MPI_Comm_dup'] = lambda n: 'MPI_Comm com[size];'
operation['MPI_Comm_dup'] = lambda n: 'MPI_Comm_dup(MPI_COMM_WORLD, &com[j]);'
error['MPI_Comm_dup'] = 'CommunicatorLeak'
fini['MPI_Comm_dup'] = lambda n: "MPI_Comm_free(&com[j]);"
free['MPI_Comm_dup'] = lambda n: ""

init['MPI_Type_contiguous'] = lambda n: 'MPI_Datatype type[size];'
operation['MPI_Type_contiguous'] = lambda n: 'MPI_Type_contiguous(2, MPI_DOUBLE, &type[j]);'
error['MPI_Type_contiguous'] = 'TypeLeak'
fini['MPI_Type_contiguous'] = lambda n: "MPI_Type_free(&type[j]);"
free['MPI_Type_contiguous'] = lambda n: ""




### P2P:basic

init['MPI_Send'] = lambda n: f'int buf{n}=rank;'
start['MPI_Send'] = lambda n: ""
operation['MPI_Send'] = lambda n: f'MPI_Send(&buf{n}, buff_size, type, dest, stag, newcom);'
fini['MPI_Send'] = lambda n: ""
free['MPI_Send'] = lambda n: ""
write['MPI_Send'] = lambda n: ""

init['MPI_Ssend'] = lambda n: f'int buf{n}=rank;'
start['MPI_Ssend'] = lambda n: ""
operation['MPI_Ssend'] = lambda n: f'MPI_Ssend(&buf{n}, buff_size, type, dest, stag, newcom);'
fini['MPI_Ssend'] = lambda n: ""
free['MPI_Ssend'] = lambda n: ""
write['MPI_Ssend'] = lambda n: ""

init['MPI_Bsend'] = lambda n: (f'int buf{n}=rank;\n'
            + f'int buffer_attached_size{n} = MPI_BSEND_OVERHEAD + sizeof(int);\n'
            + f'char* buffer_attached{n} = (char*)malloc(buffer_attached_size{n});\n'
            + f'MPI_Buffer_attach(buffer_attached{n}, buffer_attached_size{n});')
start['MPI_Bsend'] = lambda n: ""
operation['MPI_Bsend'] = lambda n: f'MPI_Bsend(&buf{n}, buff_size, type, dest, stag, newcom);'
fini['MPI_Bsend'] = lambda n: ""
free['MPI_Bsend'] = (lambda n: f'MPI_Buffer_detach(&buffer_attached{n}, &buffer_attached_size{n});\n'
            + f'free(buffer_attached{n});')
write['MPI_Bsend'] = lambda n: ""

init['MPI_Recv'] = lambda n: f'int buf{n}=-1; MPI_Status sta{n};'
start['MPI_Recv'] = lambda n: ""
operation['MPI_Recv'] = lambda n: f'MPI_Recv(&buf{n}, buff_size, type, src, rtag, newcom, &sta{n});'
fini['MPI_Recv'] = lambda n: ""
free['MPI_Recv'] = lambda n: ""
write['MPI_Recv'] = lambda n: ""

init['MPI_Probe'] = lambda n: ""
start['MPI_Probe'] = lambda n: ""
operation['MPI_Probe'] = lambda n: 'MPI_Probe(src, 0, newcom, &sta);'
fini['MPI_Probe'] = lambda n: ""
free['MPI_Probe'] = lambda n: ""
write['MPI_Probe'] = lambda n: ""



### P2P:nonblocking

init['MPI_Isend'] = lambda n: f'int buf{n}=rank; MPI_Request req{n}=MPI_REQUEST_NULL;'
start['MPI_Isend'] = lambda n: ""
operation['MPI_Isend'] = lambda n: f'MPI_Isend(&buf{n}, buff_size, type, dest, stag, newcom, &req{n});'
fini['MPI_Isend'] = lambda n: f'MPI_Wait(&req{n}, MPI_STATUS_IGNORE);'
free['MPI_Isend'] = lambda n: f'if(req{n} != MPI_REQUEST_NULL) MPI_Request_free(&req{n});'
write['MPI_Isend'] = lambda n: f'buf{n}=4;'

init['MPI_Irecv'] = lambda n: f'int buf{n}=-1; MPI_Request req{n}=MPI_REQUEST_NULL;'
start['MPI_Irecv'] = lambda n: ""
operation['MPI_Irecv'] = lambda n: f'MPI_Irecv(&buf{n}, buff_size, type, src, rtag, newcom, &req{n});'
fini['MPI_Irecv'] = lambda n: f' MPI_Wait(&req{n}, MPI_STATUS_IGNORE);'
free['MPI_Irecv'] = lambda n: f'if(req{n} != MPI_REQUEST_NULL) MPI_Request_free(&req{n});'
write['MPI_Irecv'] = lambda n: f'buf{n}++;'

### P2P:persistent

init['MPI_Send_init'] = lambda n: f'int buf{n}=rank; MPI_Request req{n}=MPI_REQUEST_NULL;'
operation['MPI_Send_init'] = lambda n: f'MPI_Send_init(&buf{n}, buff_size, type, dest, stag, newcom, &req{n});'
start['MPI_Send_init'] = lambda n: f'MPI_Start(&req{n});'
fini['MPI_Send_init'] = lambda n: f'MPI_Wait(&req{n}, MPI_STATUS_IGNORE);'
free['MPI_Send_init'] = lambda n: f'if(req{n} != MPI_REQUEST_NULL) MPI_Request_free(&req{n});'
write['MPI_Send_init'] = lambda n: f'buf{n}=4;'

init['MPI_Recv_init'] = lambda n: f'int buf{n}=-1; MPI_Request req{n}=MPI_REQUEST_NULL;'
start['MPI_Recv_init'] = lambda n: f'MPI_Start(&req{n});'
operation['MPI_Recv_init'] = lambda n: f'MPI_Recv_init(&buf{n}, buff_size, type, src, rtag, newcom, &req{n});'
fini['MPI_Recv_init'] = lambda n: f'MPI_Wait(&req{n}, MPI_STATUS_IGNORE);'
free['MPI_Recv_init'] = lambda n: f'if(req{n} != MPI_REQUEST_NULL) MPI_Request_free(&req{n});'
write['MPI_Recv_init'] = lambda n: f'buf{n}++;'

### RMA

epoch['MPI_Win_fence'] = lambda n: 'MPI_Win_fence(0, win);'
finEpoch['MPI_Win_fence'] = lambda n: 'MPI_Win_fence(0, win);'
epoch['MPI_Win_lock'] = lambda n: 'MPI_Win_lock(MPI_LOCK_SHARED, target, 0, win);'
finEpoch['MPI_Win_lock'] = lambda n: 'MPI_Win_unlock(target, win);'
epoch['MPI_Win_lock_all'] = lambda n: 'MPI_Win_lock_all(0,win);'
finEpoch['MPI_Win_lock_all'] = lambda n: 'MPI_Win_unlock_all(win);'

init['MPI_Put'] = lambda n: f'int localbuf{n}[N] = {{0}};\n  localbuf{n}[0] = 12345;'
operation['MPI_Put'] = lambda n: f'MPI_Put(&localbuf{n}, N, MPI_INT, target, 0, N, type, win);'

init['MPI_Get'] = lambda n: f'int localbuf{n}[N] = {{0}};\n  localbuf{n}[0] = 54321;'
operation['MPI_Get'] = lambda n: f'MPI_Get(&localbuf{n}, N, MPI_INT, target, 0, N, type, win);'

init['store'] = lambda n: f'int localbuf{n}[N] = {{0}};'
operation['store'] = lambda n: f'localbuf{n}[0] = 8;'

init['rstore'] = lambda n: ""
operation['rstore'] = lambda n: 'winbuf[20] = 12346;'

init['load'] = lambda n: f'int localbuf{n}[N] = {{0}};'
operation['load'] = lambda n: f'int load = localbuf{n}[0];'

init['rload'] = lambda n: ""
operation['rload'] = lambda n: "int load = winbuf[20];"

init['loadstore'] = lambda n: f'int localbuf{n}[N] = {{0}};'
operation['loadstore'] = lambda n: f'if (localbuf{n}[0] % 2 == 0)  localbuf{n}[0]++; '




def find_line(content, target, filename):
    res = 1
    for line in content.split('\n'):
        if re.search(f'[^:]{target}', line):
            #print(f'Found {target} at {line}')
            return res
        res += 1
    raise ValueError(f"Line target {target} not found in {filename}.")


def make_file(template, filename, replace):
    output = template
    filename = filename.replace("_MPI_", "_")
    replace['filename'] = filename
    # Replace all variables that don't have a ':' in their name
    while re.search(r'@\{[^@:]*\}@', output):
        m = re.search(r'@\{([^@:]*)\}@', output)
        target = m.group(1)
        #print(f"Replace @{{{target}}}@")
        if target in replace.keys():
            output = re.sub(fr'@\{{{target}\}}@', replace[target], output)
            #print(f"Replace {target} -> {replace[target]}")
        else:
            raise ValueError(f"Variable {target} used in template, but not defined.")
    # Now replace all variables with a ':' in their name: line targets are like that, and we don't want to resolve them before the others change the lines
    while re.search(r'@\{([^:@]*):([^@]*)\}@', output):
        m = re.search(r'@\{([^:@]*):([^@]*)\}@', output)
        (kind, target) = (m.group(1), m.group(2))
        if kind == 'line':
            replace = f'{find_line(output, target, filename)}'
            #print(f"Replace @{{line:{target}}}@ with '{replace}'")
            output = re.sub(fr'@\{{line:{target}\}}@', replace, output)
        else:
            raise ValueError(f"Unknown variable kind: {kind}:{target}")

    if os.path.exists(filename):
        with open(filename, 'r') as file:
            prev = file.read().split('\n')[0]
            prev = re.sub('^.*?scripts/', 'scripts/', prev)
            prev = re.sub('. DO NOT EDIT.', '', prev)
        now = output.split('\n')[0]
        now = re.sub('^.*?scripts/', 'scripts/', now)
        now = re.sub('. DO NOT EDIT.', '', now)

        print(f'WARNING: overwriting {filename}. Previously generated by: {prev}; regenerated by {now}')

    # Ready to output it
    with open(filename, 'w') as outfile:
        outfile.write(output)



## Copyright (c) 2007-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
This example shows how to work with the state profile of a host or a link,
specifying when the resource must be turned on or off.

To set such a profile, the first way is to use a file in the XML, while the second is to use the programmatic
interface, as exemplified in the main() below. Once this profile is in place, the resource will automatically
be turned on and off.

The actors running on a host that is turned off are forcefully killed
once their on_exit callbacks are executed. They cannot avoid this fate.
Since we specified on_failure="RESTART" for each actors in the XML file,
they will be automatically restarted when the host starts again.

Communications using failed links will .. fail.
"""

import sys
from simgrid import Actor, Engine, Host, Mailbox, this_actor, NetworkFailureException, TimeoutException

def master(* args):
    assert len(args) == 4, f"Actor master requires 4 parameters, but got {len(args)} ones."
    tasks_count = int(args[0])
    comp_size = int(args[1])
    comm_size = int(args[2])
    workers_count = int(args[3])

    this_actor.info(f"Got {workers_count} workers and {tasks_count} tasks to process")

    for i in range(tasks_count): # For each task to be executed:
        # - Select a worker in a round-robin way
        mailbox = Mailbox.by_name(f"worker-{i % workers_count}")
        try:
            this_actor.info(f"Send a message to {mailbox.name}")
            mailbox.put(comp_size, comm_size, 10.0)
            this_actor.info(f"Send to {mailbox.name} completed")
        except TimeoutException:
            this_actor.info(f"Mmh. Got timeouted while speaking to '{mailbox.name}'. Nevermind. Let's keep going!")
        except NetworkFailureException:
            this_actor.info(f"Mmh. The communication with '{mailbox.name}' failed. Nevermind. Let's keep going!")

    this_actor.info("All tasks have been dispatched. Let's tell everybody the computation is over.")
    for i in range(workers_count):
        # - Eventually tell all the workers to stop by sending a "finalize" task
        mailbox = Mailbox.by_name(f"worker-{i % workers_count}")
        try:
            mailbox.put(-1.0, 0, 1.0)
        except TimeoutException:
            this_actor.info(f"Mmh. Got timeouted while speaking to '{mailbox.name}'. Nevermind. Let's keep going!")
        except NetworkFailureException:
            this_actor.info(f"Mmh. The communication with '{mailbox.name}' failed. Nevermind. Let's keep going!")

    this_actor.info("Goodbye now!")

def worker(* args):
    assert len(args) == 1, "Expecting one parameter"
    my_id = int(args[0])

    mailbox = Mailbox.by_name(f"worker-{my_id}")
    done = False
    while not done:
        try:
            this_actor.info(f"Waiting a message on {mailbox.name}")
            compute_cost = mailbox.get()
            if compute_cost > 0: # If compute_cost is valid, execute a computation of that cost
                this_actor.info("Start execution...")
                this_actor.execute(compute_cost)
                this_actor.info("Execution complete.")
            else: # Stop when receiving an invalid compute_cost
                this_actor.info("I'm done. See you!")
                done = True
        except NetworkFailureException:
            this_actor.info("Mmh. Something went wrong. Nevermind. Let's keep going!")

def sleeper():
    this_actor.info("Start sleeping...")
    this_actor.sleep_for(1)
    this_actor.info("done sleeping.")

if __name__ == '__main__':
    assert len(sys.argv) > 2, f"Usage: python app-masterworkers.py platform_file deployment_file"

    e = Engine(sys.argv)

    # This is how to attach a profile to an host that is created from the XML file.
    # This should be done before calling load_platform(), as the on_creation() event is fired when loading the platform.
    # You can never set a new profile to a resource that already have one.
    def on_creation(host):
        if host.name == "Bourrassa":
            host.set_state_profile("67 0\n70 1\n", 0)
    Host.on_creation_cb(on_creation)

    e.load_platform(sys.argv[1])

    e.register_actor("master", master)
    e.register_actor("worker", worker)
    e.load_deployment(sys.argv[2])

    # Add a new host programatically, and attach a state profile to it
    lili = e.netzone_root.create_host("Lilibeth", 1e15)
    lili.set_state_profile("4 0\n5 1\n", 10)
    lili.seal()

    # Create an actor on that new host, to monitor its own state
    actor = Actor.create("sleeper", lili, sleeper)
    actor.set_auto_restart(True)

    e.run()

    this_actor.info(f"Simulation time {e.clock:.4f}")



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from argparse import ArgumentParser
import sys

from simgrid import Engine, Actor, Mailbox, this_actor


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        '--platform',
        type=str,
        required=True,
        help='path to the platform description'
    )
    return parser


def pinger(mailbox_in: Mailbox, mailbox_out: Mailbox):
    this_actor.info(f"Ping from mailbox {mailbox_in.name} to mailbox {mailbox_out.name}")

    # Do the ping with a 1-Byte payload (latency bound) ...
    payload = Engine.clock
    mailbox_out.put(payload, 1)

    # ... then wait for the (large) pong
    sender_time: float = mailbox_in.get()
    communication_time = Engine.clock - sender_time
    this_actor.info("Payload received : large communication (bandwidth bound)")
    this_actor.info(f"Pong time (bandwidth bound): {communication_time:.3f}")


def ponger(mailbox_in: Mailbox, mailbox_out: Mailbox):
    this_actor.info(f"Pong from mailbox {mailbox_in.name} to mailbox {mailbox_out.name}")

    # Receive the (small) ping first ...
    sender_time: float = mailbox_in.get()
    communication_time = Engine.clock - sender_time
    this_actor.info("Payload received : small communication (latency bound)")
    this_actor.info(f"Ping time (latency bound) {communication_time:.3f}")

    # ... Then send a 1GB pong back (bandwidth bound)
    payload = Engine.clock
    this_actor.info(f"payload = {payload:.3f}")
    mailbox_out.put(payload, int(1e9))


def main():
    settings = create_parser().parse_known_args()[0]
    e = Engine(sys.argv)
    e.load_platform(settings.platform)

    mb1: Mailbox = e.mailbox_by_name_or_create("Mailbox 1")
    mb2: Mailbox = e.mailbox_by_name_or_create("Mailbox 2")

    Actor.create("pinger", e.host_by_name("Tremblay"), pinger, mb1, mb2)
    Actor.create("ponger", e.host_by_name("Jupiter"), ponger, mb2, mb1)

    e.run()

    this_actor.info(f"Total simulation time: {e.clock:.3f}")


if __name__ == "__main__":
    main()



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
This example shows how to block on the completion of a set of communications.

As for the other asynchronous examples, the sender initiate all the messages it wants to send and
pack the resulting simgrid.Comm objects in a list. All messages thus occur concurrently.

The sender then loops until there is no ongoing communication. Using wait_any() ensures that the sender
will notice events as soon as they occur even if it does not follow the order of the container.

Here, finalize messages will terminate earlier because their size is 0, so they travel faster than the
other messages of this application.  As expected, the trace shows that the finalize of worker 1 is
processed before 'Message 5' that is sent to worker 0.
"""

import sys
from simgrid import Actor, Comm, Engine, Host, Mailbox, this_actor


FINALIZE_TAG = "finalize"


def sender(receiver_id: str, messages_count: int, payload_size: int) -> None:
    # List in which we store all ongoing communications
    pending_comms = []
    mbox = Mailbox.by_name(receiver_id)

    # Asynchronously send `messages_count` message(s) to the receiver
    for i in range(0, messages_count):
        payload = "Message {:d}".format(i)
        this_actor.info("Send '{:s}' to '{:s}'".format(payload, receiver_id))

        # Create a communication representing the ongoing communication, and store it in pending_comms
        comm = mbox.put_async(payload, payload_size)
        pending_comms.append(comm)

    # Send the final message to the receiver
    payload = FINALIZE_TAG
    final_payload_size = 0
    final_comm = mbox.put_async(payload, final_payload_size)
    pending_comms.append(final_comm)
    this_actor.info("Send '{:s}' to '{:s}".format(payload, receiver_id))
    this_actor.info("Done dispatching all messages")

    this_actor.info("Waiting for all outstanding communications to complete")
    while pending_comms:
        current_comm: Comm = pending_comms[-1]
        current_comm.wait_for(1.0)
        pending_comms.pop()
    this_actor.info("Goodbye now!")


def receiver(identifier: str) -> None:
    mbox: Mailbox = Mailbox.by_name(identifier)
    this_actor.info("Wait for my first message")
    while True:
        received = mbox.get()
        this_actor.info("I got a '{:s}'.".format(received))
        if received == FINALIZE_TAG:
            break
    this_actor.info("Goodbye now!")


def main():
    e = Engine(sys.argv)

    # Load the platform description
    e.load_platform(sys.argv[1])

    receiver_id = "receiver-0"
    Actor.create("sender", Host.by_name("Tremblay"), sender, receiver_id, 3, int(5e7))
    Actor.create("receiver", Host.by_name("Ruby"), receiver, receiver_id)

    e.run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2013-2022. The SimGrid Team.
# All rights reserved.

# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

# grep -ohrI 'bw=".*"' . | sort | uniq

import sys
import fnmatch
import os
from decimal import Decimal, DecimalException
import re


def to_str(dec):
    return re.sub(r"(\.\d*?)0*$", r"\1", dec.to_eng_string()).rstrip(".")


def convert(xml, formats, attrib):
    res = []
    m = re.search(r'%s="(.*?)"' % attrib, xml)
    while m:
        b, e = m.span(1)
        res.append(xml[:b])
        val = xml[b:e]
        xml = xml[e:]
        try:
            power = Decimal(val)
            tmp = to_str(power)
            for p, f in formats:
                d = power / p
                if d >= 1.0:
                    tmp = "%s%s" % (to_str(d), f)
                    break
            res.append(tmp)
        except DecimalException:
            print "Error with:", val
            res.append(val)
        m = re.search(r'%s="(.*?)"' % attrib, xml)

    res.append(xml)
    return "".join(res)


def formats(ll):
    return sorted(((Decimal(i), j) for i, j in ll), key=lambda x: x[0], reverse=True)


for root, dirnames, filenames in os.walk(sys.argv[1]):
    for filename in fnmatch.filter(filenames, '*.xml'):
        print root, dirnames, filename
        path = os.path.join(root, filename)
        xml = open(path).read()

        power_formats = formats([("1E0", "f"),
                                 ("1E3", "kf"),
                                 ("1E6", "Mf"),
                                 ("1E9", "Gf"),
                                 ("1E12", "Tf"),
                                 ("1E15", "Pt"),
                                 ("1E18", "Ef"),
                                 ("1E21", "Zf")])
        xml = convert(xml, power_formats, "power")

        bandwidth_formats = formats([("1E0", "Bps"),
                                     ("1E3", "kBps"),
                                     ("1E6", "MBps"),
                                     ("1E9", "GBps"),
                                     ("1E12", "TBps")])
        xml = convert(xml, bandwidth_formats, "bandwidth")
        xml = convert(xml, bandwidth_formats, "bw")
        xml = convert(xml, bandwidth_formats, "bb_bw")
        xml = convert(xml, bandwidth_formats, "bw_in")
        xml = convert(xml, bandwidth_formats, "bw_out")

        time_formats = formats([("1E-12", "ps"),
                                ("1E-9", "ns"),
                                ("1E-6", "us"),
                                ("1E-3", "ms"),
                                ("1E0", "s"),
                                ("60E0", "m"),
                                ("3600E0", "h"),
                                ("86400E0", "d"),
                                ("604800E0", "w")])
        xml = convert(xml, time_formats, "latency")
        xml = convert(xml, time_formats, "lat")
        xml = convert(xml, time_formats, "bb_lat")

        # print xml
        outfile = open(path, "w")
        outfile.write(xml)
        outfile.close()



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import subprocess

# Search for our extensions too
import sys
sys.path.append(os.path.abspath('_ext'))

# -- Run doxygen on readthedocs.org ------------------------------------------

read_the_docs_build = os.environ.get('READTHEDOCS', None) == 'True'

if read_the_docs_build:
    subprocess.call('pwd', shell=True) # should be in docs/source
    subprocess.call('doxygen', shell=True)
    subprocess.call('javasphinx-apidoc --force -o java/ ../../src/bindings/java/org/simgrid/msg', shell=True)
    subprocess.call('rm java/packages.rst', shell=True)

# -- Project information -----------------------------------------------------

project = u'SimGrid'
copyright = u'2002-2022, The SimGrid Team'
author = u'The SimGrid Team'

# The short X.Y version
version = u'3.32'

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.todo',
    'breathe',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.autosummary',
    'sphinx_tabs.tabs',
    'sphinx_copybutton',
    'javasphinx',
    'showfile',
]

todo_include_todos = True

# Setup the breath extension
breathe_projects = {'simgrid': '../build/xml'}
breathe_default_project = "simgrid"

# Setup the copybutton extension
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# Generate a warning for all a cross-reference (such as :func:`myfunc`) that cannot be found
nitpicky = True
nitpick_ignore = [
  ('cpp:identifier', 'boost'),
  ('cpp:identifier', 'boost::intrusive_ptr<Activity>'),
  ('cpp:identifier', 'boost::intrusive_ptr<Actor>'),
  ('cpp:identifier', 'boost::intrusive_ptr<Barrier>'),
  ('cpp:identifier', 'boost::intrusive_ptr<Comm>'),
  ('cpp:identifier', 'boost::intrusive_ptr<ConditionVariable>'),
  ('cpp:identifier', 'boost::intrusive_ptr<Exec>'),
  ('cpp:identifier', 'boost::intrusive_ptr<Io>'),
  ('cpp:identifier', 'boost::intrusive_ptr<Mutex>'),
  ('cpp:identifier', 'boost::intrusive_ptr<Semaphore>'),
  ('cpp:identifier', 'kernel'),
  ('cpp:identifier', 'kernel::activity'),
  ('cpp:identifier', 'kernel::activity::ActivityImplPtr'),
  ('cpp:identifier', 'kernel::activity::CommImpl'),
  ('cpp:identifier', 'kernel::activity::CommImplPtr'),
  ('cpp:identifier', 'kernel::actor'),
  ('cpp:identifier', 'kernel::actor::ActorCodeFactory'),
  ('cpp:identifier', 'kernel::actor::ActorImpl'),
  ('cpp:identifier', 'kernel::profile'),
  ('cpp:identifier', 'kernel::profile::Profile'),
  ('cpp:identifier', 'kernel::resource'),
  ('cpp:identifier', 'kernel::resource::Action'),
  ('cpp:identifier', 'kernel::resource::Action::State'),
  ('cpp:identifier', 'kernel::resource::LinkImpl'),
  ('cpp:identifier', 'kernel::resource::NetworkAction'),
  ('cpp:identifier', 'kernel::resource::SplitDuplexLinkImpl'),
  ('cpp:identifier', 'kernel::resource::StandardLinkImpl'),
  ('cpp:identifier', 'kernel::routing'),
  ('cpp:identifier', 'kernel::routing::NetPoint'),
  ('cpp:identifier', 'ns3'),
  ('cpp:identifier', 'ns3::Ptr<ns3::Node>'),
  ('cpp:identifier', 'ns3::Node'),
  ('cpp:identifier', 's4u'),
  ('cpp:identifier', 's4u_Actor'),
  ('cpp:identifier', 's4u_Barrier'),
  ('cpp:identifier', 's4u_Comm'),
  ('cpp:identifier', 's4u_ConditionVariable'),
  ('cpp:identifier', 's4u_Disk'),
  ('cpp:identifier', 's4u_Exec'),
  ('cpp:identifier', 's4u_File'),
  ('cpp:identifier', 's4u_Host'),
  ('cpp:identifier', 's4u_Link'),
  ('cpp:identifier', 's4u_Mailbox'),
  ('cpp:identifier', 's4u_Mutex'),
  ('cpp:identifier', 's4u_NetZone'),
  ('cpp:identifier', 's4u_Semaphore'),
  ('cpp:identifier', 's4u_VM'),
  ('cpp:identifier', 's4u_VirtualMachine'),
  ('cpp:identifier', 'sg_msg_Comm'),
  ('cpp:identifier', 'sg_msg_Task'),
  ('cpp:identifier', 'simgrid'),
  ('cpp:identifier', 'simgrid::s4u'),
  ('cpp:identifier', 'simgrid::s4u::Activity_T<Comm>'),
  ('cpp:identifier', 'simgrid::s4u::Activity_T<Exec>'),
  ('cpp:identifier', 'simgrid::s4u::Activity_T<Io>'),
  ('cpp:identifier', 'simgrid::s4u::this_actor'),
  ('cpp:identifier', 'simgrid::xbt'),
  ('cpp:identifier', 'simgrid::xbt::string'),
  ('cpp:identifier', 'size_t'),
  ('cpp:identifier', 'ssize_t'),
  ('cpp:identifier', 'this_actor'),
  ('cpp:identifier', 'uint64_t'),
  ('cpp:identifier', 'xbt'),
  ('cpp:identifier', 'xbt_dynar_s'),
  ('cpp:identifier', 'xbt::Extendable<Activity>'),
  ('cpp:identifier', 'xbt::Extendable<Actor>'),
  ('cpp:identifier', 'xbt::Extendable<Disk>'),
  ('cpp:identifier', 'xbt::Extendable<File>'),
  ('cpp:identifier', 'xbt::Extendable<Host>'),
  ('cpp:identifier', 'xbt::Extendable<Link>'),
  ('cpp:identifier', 'xbt::signal<void()>'),
  ('cpp:identifier', 'xbt::signal<void(Actor const&)>'),
  ('cpp:identifier', 'xbt::signal<void(Actor&)>'),
  ('cpp:identifier', 'xbt::signal<void(Comm const&)>'),
  ('cpp:identifier', 'xbt::signal<void(Comm const&, bool is_sender)>'),
  ('cpp:identifier', 'xbt::signal<void(Disk const&)>'),
  ('cpp:identifier', 'xbt::signal<void(Disk&)>'),
  ('cpp:identifier', 'xbt::signal<void(Exec const&)>'),
  ('cpp:identifier', 'xbt::signal<void(Host const&)>'),
  ('cpp:identifier', 'xbt::signal<void(Host&)>'),
  ('cpp:identifier', 'xbt::signal<void(Link const&)>'),
  ('cpp:identifier', 'xbt::signal<void(Link&)>'),
  ('cpp:identifier', 'xbt::signal<void(NetZone const&)>'),
  ('cpp:identifier', 'xbt::signal<void(VirtualMachine const&)>'),
  ('cpp:identifier', 'xbt::signal<void(const Actor&, const Host &previous_location)>'),
  ('cpp:identifier', 'xbt::signal<void(double)>'),
  ('cpp:identifier', 'xbt::signal<void(kernel::resource::NetworkAction&)>'),
  ('cpp:identifier', 'xbt::signal<void(kernel::resource::NetworkAction&, kernel::resource::Action::State)>'),
  ('cpp:identifier', 'xbt::signal<void(void)>'),
  ('cpp:identifier', 'xbt::string'),
]

# For cross-ref generation
primary_domain = 'cpp'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string: ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    'navigation_depth': 4,
    'sticky_navigation': True,
    'display_version': True,
    'includehidden': True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {'**': ['localtoc.html', 'relations.html', 'searchbox.html']}

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'simgrid-doc'

# -- Options for GitLab integration ------------------------------------------

html_context = {
    "display_gitlab": True,  # Integrate Gitlab
    "gitlab_host": "framagit.org",
    "gitlab_user": "simgrid",
    "gitlab_repo": "simgrid",
    "gitlab_version": "master",  # Version
    "conf_py_path": "/docs/source/",  # Path in the checkout to the docs root
}

html_css_files = [
    'css/custom.css',
]

# -- Other options



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from argparse import ArgumentParser
import sys

from simgrid import Engine, Actor, Comm, Mailbox, this_actor


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        '--platform',
        type=str,
        required=True,
        help='path to the platform description'
    )
    return parser


def sender(mailbox: Mailbox):
    this_actor.info("Send at full bandwidth")

    # First send a 2.5e8 Bytes payload at full bandwidth (1.25e8 Bps)
    payload = Engine.clock
    mailbox.put(payload, int(2.5e8))

    this_actor.info("Throttle the bandwidth at the Comm level")
    # ... then send it again but throttle the Comm
    payload = Engine.clock
    # get a handler on the comm first
    comm: Comm = mailbox.put_init(payload, int(2.5e8))

    # let throttle the communication. It amounts to set the rate of the comm to half the nominal bandwidth of the link,
    # i.e., 1.25e8 / 2. This second communication will thus take approximately twice as long as the first one
    comm.set_rate(int(1.25e8 / 2)).wait()


def receiver(mailbox: Mailbox):
    # Receive the first payload sent at full bandwidth
    sender_time = mailbox.get()
    communication_time = Engine.clock - sender_time
    this_actor.info(f"Payload received (full bandwidth) in {communication_time} seconds")

    # ... Then receive the second payload sent with a throttled Comm
    sender_time = mailbox.get()
    communication_time = Engine.clock - sender_time
    this_actor.info(f"Payload received (throttled) in {communication_time} seconds")


def main():
    settings = create_parser().parse_known_args()[0]
    e = Engine(sys.argv)
    e.load_platform(settings.platform)

    mailbox = e.mailbox_by_name_or_create("Mailbox")

    Actor.create("sender", e.host_by_name("node-0.simgrid.org"), sender, mailbox)
    Actor.create("receiver", e.host_by_name("node-1.simgrid.org"), receiver, mailbox)

    e.run()


if __name__ == "__main__":
    main()



# Copyright (c) 2017-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
This example demonstrate the actor migrations.

The worker actor first move by itself, and then start an execution.
During that execution, the monitor migrates the worker, that wakes up on another host.
The execution was of the right amount of flops to take exactly 5 seconds on the first host
and 5 other seconds on the second one, so it stops after 10 seconds.

Then another migration is done by the monitor while the worker is suspended.

Note that worker() takes an uncommon set of parameters,
and that this is perfectly accepted by create().
"""

import sys
from simgrid import Actor, Engine, Host, this_actor


def worker(first_host, second_host):
    flop_amount = first_host.speed * 5 + second_host.speed * 5

    this_actor.info("Let's move to {:s} to execute {:.2f} Mflops (5sec on {:s} and 5sec on {:s})".format(
        first_host.name, flop_amount / 1e6, first_host.name, second_host.name))

    this_actor.set_host(first_host)
    this_actor.execute(flop_amount)

    this_actor.info("I wake up on {:s}. Let's suspend a bit".format(
        this_actor.get_host().name))

    this_actor.suspend()

    this_actor.info("I wake up on {:s}".format(this_actor.get_host().name))
    this_actor.info("Done")


def monitor():
    boivin = Host.by_name("Boivin")
    jacquelin = Host.by_name("Jacquelin")
    fafard = Host.by_name("Fafard")

    actor = Actor.create("worker", fafard, worker, boivin, jacquelin)

    this_actor.sleep_for(5)

    this_actor.info(
        "After 5 seconds, move the actor to {:s}".format(jacquelin.name))
    actor.host = jacquelin

    this_actor.sleep_until(15)
    this_actor.info(
        "At t=15, move the actor to {:s} and resume it.".format(fafard.name))
    actor.host = fafard
    actor.resume()


if __name__ == '__main__':
    e = Engine(sys.argv)
    if len(sys.argv) < 2:
        raise AssertionError(
            "Usage: actor-migration.py platform_file [other parameters]")
    e.load_platform(sys.argv[1])

    Actor.create("monitor", Host.by_name("Boivin"), monitor)
    e.run()



# Copyright (c) 2007-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
This Python file acts as the foil to the corresponding XML file, where the
action takes place: Actors are started and stopped at predefined time
"""

import sys
from simgrid import Engine, this_actor


class Sleeper:
    """This actor just sleeps until termination"""

    def __init__(self):
        this_actor.on_exit(lambda killed: this_actor.info("Exiting now (killed)." if killed else "Exiting now (finishing)."))

    def __call__(self):
        this_actor.info("Hello! I go to sleep.")
        this_actor.sleep_for(10)
        this_actor.info("Done sleeping.")


if __name__ == '__main__':
    e = Engine(sys.argv)
    if len(sys.argv) < 2:
        raise AssertionError(
            "Usage: actor-lifetime.py platform_file actor-lifetime_d.xml [other parameters]")

    e.load_platform(sys.argv[1])     # Load the platform description
    e.register_actor("sleeper", Sleeper)
    # Deploy the sleeper actors with explicit start/kill times
    e.load_deployment(sys.argv[2])

    e.run()



#
# Copyright 2012-2015 Bronto Software, Inc. and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from setuptools import setup

setup(
    name = "javasphinx",
    packages = ["javasphinx"],
    version = "0.9.15",
    author = "Chris Thunes",
    author_email = "cthunes@brewtab.com",
    url = "http://github.com/bronto/javasphinx",
    description = "Sphinx extension for documenting Java projects",
    license = "Apache 2.0",
    classifiers = [
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries"
        ],
    install_requires=[
        "javalang>=0.10.1",
        "lxml",
        "beautifulsoup4",
        "future",
        "docutils",
        "sphinx"
    ],
    entry_points={
        'console_scripts': [
            'javasphinx-apidoc = javasphinx.apidoc:main'
            ]
        },
    long_description = """\
==========
javasphinx
==========

javasphinx is an extension to the Sphinx documentation system which adds support
for documenting Java projects. It includes a Java domain for writing
documentation manually and a javasphinx-apidoc utility which will automatically
generate API documentation from existing Javadoc markup.
"""
)



#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: MBI

  Description: @{shortdesc}@
    @{longdesc}@

  Version of MPI: Conforms to MPI 1.1, does not require MPI 2 implementation

BEGIN_MPI_FEATURES
  P2P!basic: Lacking
  P2P!nonblocking: Lacking
  P2P!persistent: Lacking
  COLL!basic: @{collfeature}@
  COLL!nonblocking: @{icollfeature}@
  COLL!persistent: Lacking
  COLL!tools: @{toolfeature}@
  RMA: Lacking
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 2 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

#define buff_size 128

int main(int argc, char **argv) {
  int nprocs = -1;
  int rank = -1;
  int root = 0;
  int size = 1;
  int j = 0;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (nprocs < 2)
    printf("MBI ERROR: This test needs at least 2 processes to produce a bug!\\n");

  MPI_Comm newcom = MPI_COMM_WORLD;
  MPI_Op op = MPI_SUM;
  MPI_Datatype type = MPI_INT;

  int dbs = sizeof(int)*nprocs; /* Size of the dynamic buffers for alltoall and friends */

  @{init}@
  @{start}@

  @{change_arg}@
  @{operation}@ /* MBIERROR2 */
  @{fini}@
  @{free}@

  MPI_Finalize();
  printf("Rank %d finished normally\\n", rank);
  return 0;
}
"""

#####################################################
# Generate code with color mismatch in MPI_Comm_split
#####################################################

for c in gen.tcoll4color:
    patterns = {}
    patterns = {'c': c}
    patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
    patterns['collfeature'] = 'Lacking'
    patterns['icollfeature'] = 'Lacking'
    patterns['toolfeature'] = 'Yes' if c in gen.tcoll4color else 'Lacking'
    patterns['c'] = c
    patterns['init'] = gen.init[c]("1")
    patterns['start'] = gen.start[c]("1")
    patterns['operation'] = gen.operation[c]("1")
    patterns['fini'] = gen.fini[c]("1")
    patterns['free'] = gen.free[c]("1")
    patterns['change_arg'] = ''

    # Generate the code with invalid color
    replace = patterns.copy()
    replace['shortdesc'] = 'Invalid color in @{c}@'
    replace['longdesc'] = 'invalid color in @{c}@'
    replace['outcome'] = 'ERROR: InvalidOtherArg'
    replace['errormsg'] = 'Invalid Argument in collective. @{c}@ at line @{line:MBIERROR2}@ has an invalid color'
    replace['change_arg'] = 'color=-10; /* MBIERROR1*/'
    gen.make_file(template, f'InvalidParam_OtherArg_{c}_nok.c', replace)


##################################
# Generate code with root mismatch
##################################

for c in gen.coll4root + gen.icoll4root:
    patterns = {}
    patterns = {'c': c}
    patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
    patterns['collfeature'] = 'Yes' if c in gen.coll4root else 'Lacking'
    patterns['icollfeature'] = 'Yes' if c in gen.icoll4root else 'Lacking'
    patterns['toolfeature'] = 'Lacking'
    patterns['c'] = c
    patterns['init'] = gen.init[c]("1")
    patterns['start'] = gen.start[c]("1")
    patterns['fini'] = gen.fini[c]("1")
    patterns['free'] = gen.free[c]("1")
    patterns['operation'] = gen.operation[c]("1")
    patterns['change_arg'] = ''

    # Generate an incorrect root matching (root mismatch)
    replace = patterns.copy()
    replace['shortdesc'] = 'Collective @{c}@ with a root mismatch'
    replace['longdesc'] = 'Odd ranks use 0 as a root while even ranks use 1 as a root'
    replace['outcome'] = 'ERROR: RootMatching'
    replace['errormsg'] = 'Collective root mistmatch. @{c}@ at @{filename}@:@{line:MBIERROR2}@ has 0 or 1 as a root.'
    replace['change_arg'] = 'if (rank % 2)\n    root = 1; /* MBIERROR1 */'
    gen.make_file(template, f'ParamMatching_Root_{c}_nok.c', replace)

    # Generate the call with root=-1 (invalid root)
    replace = patterns.copy()
    replace['shortdesc'] = f'Collective {c} with root = -1'
    replace['longdesc'] = f'Collective {c} with root = -1'
    replace['outcome'] = 'ERROR: InvalidRoot'
    replace['errormsg'] = 'Invalid collective root.  @{c}@ at @{filename}@:@{line:MBIERROR2}@ has -1 as a root while communicator MPI_COMM_WORLD requires ranks in range 0 to 1.'
    replace['change_arg'] = 'root = -1; /* MBIERROR1 */'
    gen.make_file(template, f'InvalidParam_RootNeg_{c}_nok.c', replace)

    # Generate the call with root=2 (root not in communicator)
    replace = patterns.copy()
    replace['shortdesc'] = f'Collective {c} with root out of the communicator'
    replace['longdesc'] = f'Collective {c} with root = 2 (there is only 2 ranks)'
    replace['outcome'] = 'ERROR: InvalidRoot'
    replace['errormsg'] = 'Invalid collective root.  @{c}@ at @{filename}@:@{line:MBIERROR2}@ has 2 as a root while communicator MPI_COMM_WORLD requires ranks in range 0 to 1.'
    replace['change_arg'] = 'root = nprocs; /* MBIERROR1 */'
    gen.make_file(template, f'InvalidParam_RootTooLarge_{c}_nok.c', replace)


##################################
# Generate code with type mismatch
##################################

for c in gen.coll + gen.icoll:
    if c != 'MPI_Barrier': # Barrier has no Data to mismatch or to nullify
        patterns = {}
        patterns = {'c': c}
        patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
        patterns['collfeature'] = 'Yes' if c in gen.coll else 'Lacking'
        patterns['icollfeature'] = 'Yes' if c in gen.icoll + gen.ibarrier else 'Lacking'
        patterns['toolfeature'] = 'Lacking'
        patterns['c'] = c
        patterns['init'] = gen.init[c]("1")
        patterns['start'] = gen.start[c]("1")
        patterns['fini'] = gen.fini[c]("1")
        patterns['operation'] = gen.operation[c]("1")
        patterns['free'] = gen.free[c]("1")
        patterns['change_arg'] = ''

        # Generate the incorrect matching (datatype Mmismatch)
        replace = patterns.copy()
        replace['shortdesc'] = 'Collective @{c}@ with a datatype mismatch'
        replace['longdesc'] = 'Odd ranks use MPI_INT as the datatype while even ranks use MPI_FLOAT'
        replace['outcome'] = 'ERROR: DatatypeMatching'
        replace['errormsg'] = 'Collective datatype mistmatch. @{c}@ at @{filename}@:@{line:MBIERROR2}@ has MPI_INT or MPI_FLOAT as a datatype.'
        replace['change_arg'] = 'if (rank % 2)\n    type = MPI_FLOAT; /* MBIERROR1 */'
        gen.make_file(template, f'ParamMatching_Data_{c}_nok.c', replace)

        # Generate the call with null type (invalid datatype)
        replace = patterns.copy()
        replace['shortdesc'] = 'Collective @{c}@ with an invalid datatype '
        replace['longdesc'] = 'Collective @{c}@ with an invalid datatype '
        replace['outcome'] = 'ERROR: InvalidDatatype'
        replace['errormsg'] = 'Invalid Datatype. @{c}@ at @{filename}@:@{line:MBIERROR2}@ has an invalid datatype.'
        replace['change_arg'] = 'type=MPI_DATATYPE_NULL; /* MBIERROR1 */'
        gen.make_file(template, f'InvalidParam_DataNull_{c}_nok.c', replace)


##################################
# Generate code with Op  mismatch
##################################

for c in gen.coll4op + gen.icoll4op:
    patterns = {}
    patterns = {'c': c}
    patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
    patterns['collfeature'] = 'Yes' if c in  gen.coll4op else 'Lacking'
    patterns['icollfeature'] = 'Yes' if c in gen.icoll4op else 'Lacking'
    patterns['toolfeature'] = 'Lacking'
    patterns['c'] = c
    patterns['init'] = gen.init[c]("1")
    patterns['start'] = gen.start[c]("1")
    patterns['fini'] = gen.fini[c]("1")
    patterns['operation'] = gen.operation[c]("1")
    patterns['free'] = gen.free[c]("1")
    patterns['change_arg'] = ''

    # Generate the incorrect matching (op mismatch)
    replace = patterns.copy()
    replace['shortdesc'] = 'Collective @{c}@ with an operator  mismatch'
    replace['longdesc'] = 'Odd ranks use MPI_SUM as the operator while even ranks use MPI_MAX'
    replace['outcome'] = 'ERROR: OperatorMatching'
    replace['errormsg'] = 'Collective operator mistmatch. @{c}@ at @{filename}@:@{line:MBIERROR2}@ has MPI_MAX or MPI_SUM as an operator.'
    replace['change_arg'] = 'if (rank % 2)\n    op = MPI_MAX; /* MBIERROR1 */'
    gen.make_file(template, f'ParamMatching_Op_{c}_nok.c', replace)

    # Generate the call with Op=MPI_OP_NULL (invalid op)
    replace = patterns.copy()
    replace['shortdesc'] = 'Collective @{c}@ with an invalid operator '
    replace['longdesc'] = 'Collective @{c}@ with an invalid operator '
    replace['outcome'] = 'ERROR: InvalidOperator'
    replace['errormsg'] = 'Invalid Operator. @{c}@ at @{filename}@:@{line:MBIERROR2}@ has MPI_OP_NULL as an operator.'
    replace['change_arg'] = 'op = MPI_OP_NULL; /* MBIERROR1 */'
    gen.make_file(template, f'InvalidParam_OpNull_{c}_nok.c', replace)



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.

#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
This example shows how to use simgrid::s4u::this_actor::wait() to wait for a given communication.

As for the other asynchronous examples, the sender initiate all the messages it wants to send and
pack the resulting simgrid::s4u::CommPtr objects in a vector. All messages thus occurs concurrently.

The sender then loops until there is no ongoing communication.
"""

import sys
from simgrid import Actor, Engine, Host, Mailbox, this_actor


def sender(messages_count, msg_size, receivers_count):
    # List in which we store all ongoing communications
    pending_comms = []

    # Vector of the used mailboxes
    mboxes = [Mailbox.by_name("receiver-{:d}".format(i)) for i in range(0, receivers_count)]

    # Start dispatching all messages to receivers, in a round robin fashion
    for i in range(0, messages_count):
        content = "Message {:d}".format(i)
        mbox = mboxes[i % receivers_count]

        this_actor.info("Send '{:s}' to '{:s}'".format(content, str(mbox)))

        # Create a communication representing the ongoing communication, and store it in pending_comms
        comm = mbox.put_async(content, msg_size)
        pending_comms.append(comm)

    # Start sending messages to let the workers know that they should stop
    for i in range(0, receivers_count):
        mbox = mboxes[i]
        this_actor.info("Send 'finalize' to '{:s}'".format(str(mbox)))
        comm = mbox.put_async("finalize", 0)
        pending_comms.append(comm)

    this_actor.info("Done dispatching all messages")

    # Now that all message exchanges were initiated, wait for their completion, in order of creation.
    for comm in pending_comms:
        comm.wait()
    this_actor.info("Goodbye now!")

def receiver(my_id):
    mbox = Mailbox.by_name("receiver-{:d}".format(my_id))
    this_actor.info("Wait for my first message")
    while True:
        received = mbox.get()
        this_actor.info("I got a '{:s}'.".format(received))
        if received == "finalize":
            break  # If it's a finalize message, we're done.


if __name__ == '__main__':
    e = Engine(sys.argv)

    e.load_platform(sys.argv[1])             # Load the platform description

    Actor.create("sender", Host.by_name("Tremblay"), sender, 3, 50000000, 1)
    Actor.create("receiver", Host.by_name("Ruby"), receiver, 0)

    e.run()



# Copyright (c) 2018-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.


# This script does exactly the same thing as file s4u-exec-ptask.cpp

import sys
from simgrid import Actor, Engine, Host, this_actor, TimeoutException

def runner():
    hosts = Engine.instance.all_hosts
    hosts_count = len(hosts)

    # Test 1
    this_actor.info("First, build a classical parallel activity, with 1 Gflop to execute on each node, "
               "and 10MB to exchange between each pair")
    computation_amounts = [1e9]*hosts_count
    communication_amounts = [0]*hosts_count*hosts_count
    for i in range(hosts_count):
        for j in range(i+1, hosts_count):
            communication_amounts[i * hosts_count + j] = 1e7
    this_actor.parallel_execute(hosts, computation_amounts, communication_amounts)

    # Test 2
    this_actor.info("We can do the same with a timeout of 10 seconds enabled.")
    activity = this_actor.exec_init(hosts, computation_amounts, communication_amounts)
    try:
        activity.wait_for(10.0)
        sys.exit("Woops, this did not timeout as expected... Please report that bug.")
    except TimeoutException:
        this_actor.info("Caught the expected timeout exception.")
        activity.cancel()

    # Test 3
    this_actor.info("Then, build a parallel activity involving only computations (of different amounts) and no communication")
    computation_amounts = [3e8, 6e8, 1e9]
    communication_amounts = []
    this_actor.parallel_execute(hosts, computation_amounts, communication_amounts)

    # Test 4
    this_actor.info("Then, build a parallel activity with no computation nor communication (synchro only)")
    computation_amounts = []
    this_actor.parallel_execute(hosts, computation_amounts, communication_amounts)

    # Test 5
    this_actor.info("Then, Monitor the execution of a parallel activity")
    computation_amounts = [1e6]*hosts_count
    communication_amounts = [0, 1e6, 0, 0, 0, 1e6, 1e6, 0, 0]
    activity = this_actor.exec_init(hosts, computation_amounts, communication_amounts)
    activity.start()
    while not activity.test():
        ratio = activity.remaining_ratio * 100
        this_actor.info(f"Remaining flop ratio: {ratio:.0f}%")
        this_actor.sleep_for(5)
    activity.wait()

    # Test 6
    this_actor.info("Finally, simulate a malleable task (a parallel execution that gets reconfigured after its start).")
    this_actor.info("  - Start a regular parallel execution, with both comm and computation")
    computation_amounts = [1e6]*hosts_count
    communication_amounts = [0, 1e6, 0, 0, 1e6, 0, 1e6, 0, 0]
    activity = this_actor.exec_init(hosts, computation_amounts, communication_amounts)
    activity.start()
    this_actor.sleep_for(10)
    remaining_ratio = activity.remaining_ratio
    this_actor.info(f"  - After 10 seconds, {remaining_ratio*100:.2f}% remains to be done. Change it from 3 hosts to 2 hosts only.")
    this_actor.info("    Let's first suspend the task.")
    activity.suspend()
    this_actor.info("  - Now, simulate the reconfiguration (modeled as a comm from the removed host to the remaining ones).")
    rescheduling_comp = [0, 0, 0]
    rescheduling_comm = [0, 0, 0, 0, 0, 0, 25000, 25000, 0]
    this_actor.parallel_execute(hosts, rescheduling_comp, rescheduling_comm)
    this_actor.info("  - Now, let's cancel the old task and create a new task with modified comm and computation vectors:")
    this_actor.info("    What was already done is removed, and the load of the removed host is shared between remaining ones.")
    for i in range(2):
        # remove what we've done so far, for both comm and compute load
        computation_amounts[i] *= remaining_ratio
        communication_amounts[i] *= remaining_ratio
        # The work from 1 must be shared between 2 remaining ones. 1/2=50% of extra work for each
        computation_amounts[i] *= 1.5
        communication_amounts[i] *= 1.5
    hosts = hosts[:2]
    computation_amounts = computation_amounts[:2]
    remaining_comm = communication_amounts[1]
    communication_amounts = [0, remaining_comm, remaining_comm, 0]
    activity.cancel()
    activity = this_actor.exec_init(hosts, computation_amounts, communication_amounts)
    this_actor.info("  - Done, let's wait for the task completion")
    activity.wait()
    this_actor.info("Goodbye now!")


if __name__ == "__main__":
    if len(sys.argv) != 2:
        sys.exit(f"Syntax: {sys.argv[0]} <platform_file>")
    platform = sys.argv[1]
    engine = Engine.instance
    Engine.set_config("host/model:ptask_L07")  # /!\ this is required for running ptasks
    engine.load_platform(platform)
    Actor.create("foo", engine.host_by_name("MyHost1"), runner)
    engine.run()



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
This simple example demonstrates the Comm.sento_init() Comm.sento_async() functions,
that can be used to create a direct communication from one host to another without
relying on the mailbox mechanism.

There is not much to say, actually: The _init variant creates the communication and
leaves it unstarted (in case you want to modify this communication before it starts),
while the _async variant creates and start it. In both cases, you need to wait() it.

It is mostly useful when you want to have a centralized simulation of your settings,
with a central actor declaring all communications occurring on your distributed system.
"""

from argparse import ArgumentParser
import sys

from simgrid import Actor, Comm, Engine, Host, this_actor


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        '--platform',
        type=str,
        required=True,
        help='path to the platform description'
    )
    return parser


def sender(h1: Host, h2: Host, h3: Host, h4: Host):
    this_actor.info(f"Send c12 with sendto_async({h1.name} -> {h2.name}),"
                    f" and c34 with sendto_init({h3.name} -> {h4.name})")
    c12: Comm = Comm.sendto_async(h1, h2, int(1.5e7))
    c34: Comm = Comm.sendto_init(h3, h4)
    c34.set_payload_size(int(1e7))

    # You can also detach() communications that you never plan to test() or wait().
    # Here we create a communication that only slows down the other ones
    noise: Comm = Comm.sendto_init(h1, h2)
    noise.set_payload_size(10000)
    noise.detach()

    this_actor.info(f"After creation, c12 is {c12.state_str} (remaining: {c12.remaining:.2e} bytes);"
                    f" c34 is {c34.state_str} (remaining: {c34.remaining:.2e} bytes)")
    this_actor.sleep_for(1)
    this_actor.info(f"One sec later, c12 is {c12.state_str} (remaining: {c12.remaining:.2e} bytes);"
                    f" c34 is {c34.state_str} (remaining: {c34.remaining:.2e} bytes)")
    c34.start()
    this_actor.info(f"After c34.start(), c12 is {c12.state_str} (remaining: {c12.remaining:.2e} bytes);"
                    f" c34 is {c34.state_str} (remaining: {c34.remaining:.2e} bytes)")
    c12.wait()
    this_actor.info(f"After c12.wait(), c12 is {c12.state_str} (remaining: {c12.remaining:.2e} bytes);"
                    f" c34 is {c34.state_str} (remaining: {c34.remaining:.2e} bytes)")
    c34.wait()
    this_actor.info(f"After c34.wait(), c12 is {c12.state_str} (remaining: {c12.remaining:.2e} bytes);"
                    f" c34 is {c34.state_str} (remaining: {c34.remaining:.2e} bytes)")

    # As usual, you don't have to explicitly start communications that were just init()ed.
    # The wait() will start it automatically.
    c14: Comm = Comm.sendto_init(h1, h4)
    c14.set_payload_size(100).wait()


def main():
    settings = create_parser().parse_known_args()[0]
    e = Engine(sys.argv)
    e.load_platform(settings.platform)
    Actor.create(
        "sender", e.host_by_name("Boivin"), sender,
        e.host_by_name("Tremblay"),  # h1
        e.host_by_name("Jupiter"),  # h2
        e.host_by_name("Fafard"),  # h3
        e.host_by_name("Ginette")  # h4
    )
    e.run()
    this_actor.info(f"Total simulation time: {e.clock:.3f}")


if __name__ == "__main__":
    main()



#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: MBI

  Description: @{shortdesc}@
    @{longdesc}@

  Version of MPI: Conforms to MPI 2, does not require MPI 3 implementation

BEGIN_MPI_FEATURES
  P2P!basic: @{p2pfeature}@
  P2P!nonblocking: @{ip2pfeature}@
  P2P!persistent: Lacking
  COLL!basic: Lacking
  COLL!nonblocking: Lacking
  COLL!persistent: Lacking
  COLL!tools: Lacking
  RMA: @{rmafeature}@
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 4 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

#define N 1

int main(int argc, char **argv) {
  int nprocs = -1;
  int rank = -1;
  MPI_Win win;
  int * winbuf = malloc(N * sizeof(int)); // Window buffer
  int buff_size = 1;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (nprocs < 4)
    printf("MBI ERROR: This test needs at least 4 processes to produce a bug!\\n");

  MPI_Comm newcom = MPI_COMM_WORLD;
  MPI_Datatype type = MPI_INT;
  int stag=0, rtag=0;
  winbuf[0] = nprocs;

  MPI_Win_create(winbuf, N*sizeof(int), sizeof(int), MPI_INFO_NULL, MPI_COMM_WORLD, &win);


  @{init1}@
  @{init2}@
  @{init3}@

  if (rank == 0) {
    int target=1;
    MPI_Win_lock(MPI_LOCK_EXCLUSIVE, 1, 0, win);
    @{operation1}@
    localbuf1[0] = 12345; /* MBIERROR1 */
    MPI_Win_unlock(1, win);
  }else if (rank == 2){
    int dest=1;
    @{operation2}@
    @{fini2}@
  }else if (rank == 1){
    int src=2;
    buf3 = winbuf[0];
    @{operation3}@
    winbuf[0] = buf3; /* MBIERROR2 */
    @{fini3}@
  }

  MPI_Win_free(&win);
  free(winbuf);

  MPI_Finalize();
  printf("Rank %d finished normally\\n", rank);
  return 0;
}
"""


for p in gen.put + gen.get:
    for s in gen.send + gen.isend:
        for r in gen.recv + gen.irecv:
            patterns = {}
            patterns = {'p': p, 's': s, 'r': r}
            patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
            patterns['rmafeature'] = 'Yes'
            patterns['p2pfeature'] = 'Yes' if s in gen.send or r in gen.recv  else 'Lacking'
            patterns['ip2pfeature'] = 'Yes' if s in gen.isend or r in gen.irecv  else 'Lacking'
            patterns['p'] = p
            patterns['s'] = s
            patterns['r'] = r
            patterns['init1'] = gen.init[p]("1")
            patterns['init2'] = gen.init[s]("2")
            patterns['init3'] = gen.init[r]("3")
            patterns['fini2'] = gen.fini[s]("2")
            patterns['fini3'] = gen.fini[r]("3")
            patterns['operation1'] = gen.operation[p]("1") #put or get
            patterns['operation2'] = gen.operation[s]("2") #send
            patterns['operation3'] = gen.operation[r]("3") #recv

            replace = patterns.copy()
            replace['shortdesc'] = 'Global Concurrency error.'
            replace['longdesc'] = 'Global Concurrency error. Concurrent access of variable winbuf by @{p}@ and @{r}@'
            replace['outcome'] = 'ERROR: GlobalConcurrency'
            replace['errormsg'] = 'Global Concurrency error. @{p}@ at @{filename}@:@{line:MBIERROR1}@ accesses the window of process 1. Process 1 receives data from process 2 and uses variable winbuf. winbuf in process 1 is then nondeterministic.'
            gen.make_file(template, f'GlobalConcurrency_{p}_{s}_{r}_nok.c', replace)



# Copyright (c) 2017-2022. The SimGrid Team. All rights reserved.
#
# This program is free software you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
Usage: actor-join.py platform_file [other parameters]
"""

import sys
from simgrid import Actor, Engine, Host, this_actor


def sleeper():
    this_actor.info("Sleeper started")
    this_actor.sleep_for(3)
    this_actor.info("I'm done. See you!")


def master():
    this_actor.info("Start 1st sleeper")
    actor = Actor.create("1st sleeper from master", Host.current(), sleeper)
    this_actor.info("Join the 1st sleeper (timeout 2)")
    actor.join(2)

    this_actor.info("Start 2nd sleeper")
    actor = Actor.create("2nd sleeper from master", Host.current(), sleeper)
    this_actor.info("Join the 2nd sleeper (timeout 4)")
    actor.join(4)

    this_actor.info("Start 3rd sleeper")
    actor = Actor.create("3rd sleeper from master", Host.current(), sleeper)
    this_actor.info("Join the 3rd sleeper (timeout 2)")
    actor.join(2)

    this_actor.info("Start 4th sleeper")
    actor = Actor.create("4th sleeper from master", Host.current(), sleeper)
    this_actor.info("Waiting 4")
    this_actor.sleep_for(4)
    this_actor.info("Join the 4th sleeper after its end (timeout 1)")
    actor.join(1)

    this_actor.info("Goodbye now!")

    this_actor.sleep_for(1)

    this_actor.info("Goodbye now!")


if __name__ == '__main__':
    e = Engine(sys.argv)
    if len(sys.argv) < 2:
        raise AssertionError(
            "Usage: actor-join.py platform_file [other parameters]")

    e.load_platform(sys.argv[1])

    Actor.create("master", Host.by_name("Tremblay"), master)

    e.run()

    this_actor.info("Simulation time {}".format(e.clock))



#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: @{origin}@

  Description: @{shortdesc}@
    @{longdesc}@

   Version of MPI: Conforms to MPI 1.1, does not require MPI 2 implementation

BEGIN_MPI_FEATURES
  P2P!basic: @{p2pfeature}@
  P2P!nonblocking: @{ip2pfeature}@
  P2P!persistent: @{persfeature}@
  COLL!basic: Lacking
  COLL!nonblocking: Lacking
  COLL!persistent: Lacking
  COLL!tools: Yes
  RMA: Lacking
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 2 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>


int main(int argc, char **argv) {
  int nprocs = -1;
  int rank = -1;
  int src=0, dest=1;
  int stag = 0, rtag = 0;
  int buff_size = 1;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (nprocs < 2)
    printf("MBI ERROR: This test needs at least 2 processes to produce a bug!\\n");

  MPI_Datatype type = MPI_INT;
  MPI_Comm newcom;
  MPI_Comm_split(MPI_COMM_WORLD, 0, nprocs - rank, &newcom);
  @{change_com}@
  @{change_srcdest}@

  @{init1}@
  @{init2}@
  if (rank == 0) {
    @{operation1}@ /* MBIERROR1 */
    @{start1}@
    @{fini1}@
  }else if (rank == 1) {
    @{operation2}@ /* MBIERROR2 */
    @{start2}@
    @{fini2}@
  }
  @{free1}@
  @{free2}@

  if(newcom != MPI_COMM_NULL && newcom != MPI_COMM_WORLD)
    MPI_Comm_free(&newcom);

  MPI_Finalize();
  printf("Rank %d finished normally\\n", rank);
  return 0;
}
"""


for p1 in gen.send + gen.isend + gen.psend:
    for p2 in gen.recv + gen.irecv + gen.precv:
        patterns = {}
        patterns = {'p1': p1, 'p2': p2}
        patterns['origin'] = "MBI"
        patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
        patterns['p2pfeature'] = 'Yes' if p1 in gen.send or p2 in gen.recv  else 'Lacking'
        patterns['ip2pfeature'] = 'Yes' if p1 in gen.isend or p2 in gen.irecv  else 'Lacking'
        patterns['persfeature'] = 'Yes' if p1 in gen.psend or p2 in gen.precv  else 'Lacking'
        patterns['p1'] = p1
        patterns['p2'] = p2
        patterns['init1'] = gen.init[p1]("1")
        patterns['init2'] = gen.init[p2]("2")
        patterns['start1'] = gen.start[p1]("1")
        patterns['start2'] = gen.start[p2]("2")
        patterns['fini1'] = gen.fini[p1]("1")
        patterns['fini2'] = gen.fini[p2]("2")
        patterns['operation1'] = gen.operation[p1]("1") #send
        patterns['operation2'] = gen.operation[p2]("2") #recv
        patterns['free1'] = gen.free[p1]("1")
        patterns['free2'] = gen.free[p2]("2")
        patterns['change_srcdest'] = ""
        patterns['change_com'] = ""

        # Generate the incorrect matching
        replace = patterns.copy()
        replace['shortdesc'] = 'Point to point @{p1}@ and @{p2}@ have a communicator mismatch'
        replace['longdesc'] = 'Process 1 uses newcom as the communicator while process 0 uses MPI_COMM_WORLD.'
        replace['outcome'] = 'ERROR: CommunicatorMatching'
        replace['errormsg'] = 'P2P Communicator mismatch. @{p1}@ at @{filename}@:@{line:MBIERROR1}@ and @{p2}@ at @{filename}@:@{line:MBIERROR2}@ have newcom or MPI_COMM_WORLD as a communicator.'
        replace['change_com'] = 'if (rank==0)\n    newcom = MPI_COMM_WORLD; /* MBIERROR */'
        gen.make_file(template, f'ParamMatching_Com_{p1}_{p2}_nok.c', replace)

        # Generate the code with an invalid communicator
        replace = patterns.copy()
        replace['shortdesc'] = 'Point to point @{p1}@ and @{p2}@ have an invalid communicator'
        replace['longdesc'] = 'Point to point @{p1}@ and @{p2}@ have an invalid communicator.'
        replace['outcome'] = 'ERROR: InvalidCommunicator'
        replace['errormsg'] = 'Invalid Communicator. @{p1}@ at @{filename}@:@{line:MBIERROR1}@ and @{p2}@ at @{filename}@:@{line:MBIERROR2}@ use a communicator that is freed line @{line:MBIERROR}@.'
        replace['change_com'] = 'MPI_Comm_free(&newcom);  /* MBIERROR */'
        gen.make_file(template, f'InvalidParam_Com_{p1}_{p2}_nok.c', replace)

        #  Generate the code with an invalid communicator ==> TO CHECK
        #replace = patterns.copy()
        #replace['shortdesc'] = 'Point to point @{p1}@ and @{p2}@ have an invalid communicator'
       # replace['longdesc'] = 'Point to point @{p1}@ and @{p2}@ have an invalid communicator.'
       # replace['outcome'] = 'ERROR: InvalidCommunicator'
       # replace['errormsg'] = 'Invalid Communicator. @{p1}@ at @{filename}@:@{line:MBIERROR1}@ and @{p2}@ at @{filename}@:@{line:MBIERROR2}@ use different communicators'
       # replace['origin'] = "MPI-Corrbench"
       # replace['change_com'] = ""
       # gen.make_file(template, f'InvalidParam_Com_{p1}_{p2}_nok.c', replace)

        # Generate the code with an invalid dest
        replace = patterns.copy()
        replace['origin'] = "MBI"
        replace['shortdesc'] = 'Point to point @{p1}@ has an invalid argument'
        replace['longdesc'] = 'Point to point @{p1}@ and @{p2}@ have an invalid communicator.'
        replace['outcome'] = 'ERROR: InvalidSrcDest'
        replace['errormsg'] = 'InvalidSrcDest. @{p1}@ at @{filename}@:@{line:MBIERROR1}@ performs a send with a dest not in communicator (dest is changed line @{line:MBIERROR}@).'
        replace['change_com'] = ""
        replace['change_srcdest'] = 'dest=4; /* MBIERROR */'
        gen.make_file(template, f'InvalidParam_Dest_{p1}_{p2}_nok.c', replace)

        # Generate the code with an invalid src
        replace = patterns.copy()
        replace['shortdesc'] = 'Point to point @{p2}@ has an invalid argument'
        replace['longdesc'] = 'Point to point @{p1}@ and @{p2}@ have an invalid communicator.'
        replace['outcome'] = 'ERROR: InvalidSrcDest'
        replace['errormsg'] = 'InvalidSrcDest. @{p2}@ at @{filename}@:@{line:MBIERROR2}@ performs a recv with a negative integer as source (src is changed line @{line:MBIERROR}@).'
        replace['change_srcdest'] = 'src=-1; /* MBIERROR */'
        gen.make_file(template, f'InvalidParam_Src_{p1}_{p2}_nok.c', replace)



#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: @{origin}@

  Description: @{shortdesc}@
    @{longdesc}@


BEGIN_MPI_FEATURES
  P2P!basic: Lacking
  P2P!nonblocking: Lacking
  P2P!persistent: Lacking
  COLL!basic: Lacking
  COLL!nonblocking: Lacking
  COLL!persistent: Lacking
  COLL!tools: Lacking
  RMA: @{rmafeature}@
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 2 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

#define N 10

int main(int argc, char **argv) {
  int rank, numProcs;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &numProcs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  int *winbuf = malloc(N * sizeof(int));

  MPI_Win win;
  MPI_Win_create(&winbuf, N * sizeof(int), 1, MPI_INFO_NULL, MPI_COMM_WORLD, &win);

  MPI_Datatype type = MPI_INT;
  int target = (rank + 1) % numProcs;

  if(rank == 0){
    @{epoch}@
    @{change_arg}@
    @{init}@
     @{operation}@ /* MBIERROR2 */

    @{finEpoch}@
  } else {
    @{epoch}@

    @{finEpoch}@
  }

  MPI_Win_free(&win);

  free(winbuf);

  MPI_Finalize();
  return 0;
}
"""


for e in gen.epoch:
    for p in gen.rma:
        patterns = {}
        patterns = {'e': e, 'p': p}
        patterns['origin'] = "MBI"
        patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
        patterns['rmafeature'] = 'Yes'
        patterns['p'] = p
        patterns['e'] = e
        patterns['epoch'] = gen.epoch[e]("1")
        patterns['finEpoch'] = gen.finEpoch[e]("1")
        patterns['init'] = gen.init[p]("1")
        patterns['operation'] = gen.operation[p]("1")
        patterns['change_arg'] = ""

        # Generate a code with a null type
        replace = patterns.copy()
        replace['shortdesc'] = 'Invalid argument in one-sided operation.'
        replace['longdesc'] = 'A one-sided operation has MPI_DATATYPE_NULL as a type.'
        replace['outcome'] = 'ERROR: InvalidDatatype'
        replace['change_arg'] = 'type = MPI_DATATYPE_NULL;'
        replace['errormsg'] = '@{p}@ at @{filename}@:@{line:MBIERROR}@ has MPI_DATATYPE_NULL as a type'
        gen.make_file(template, f'InvalidParam_BufferNullCond_{e}_{p}_nok.c', replace)

        # Generate a code with an invalid type
        replace = patterns.copy()
        replace['shortdesc'] = 'Invalid argument in one-sided operation.'
        replace['longdesc'] = 'Use of an invalid datatype in one-sided operation.'
        replace['outcome'] = 'ERROR: InvalidDatatype'
        replace['change_arg'] = 'MPI_Type_contiguous (2, MPI_INT, &type); MPI_Type_commit(&type);MPI_Type_free(&type); /* MBIERROR2 */'
        replace['errormsg'] = 'Invalid Datatype in @{p}@ at @{filename}@:@{line:MBIERROR}@'
        gen.make_file(template, f'InvalidParam_DatatypeCond_{e}_{p}_nok.c', replace)



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from argparse import ArgumentParser
from typing import List
import sys

from simgrid import Engine, Actor, Comm, Mailbox, this_actor


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        '--platform',
        type=str,
        required=True,
        help='path to the platform description'
    )
    return parser


def rank0():
    rank0_mailbox: Mailbox = Mailbox.by_name("rank0")
    this_actor.info("Post my asynchronous receives")
    comm1, a1 = rank0_mailbox.get_async()
    comm2, a2 = rank0_mailbox.get_async()
    comm3, a3 = rank0_mailbox.get_async()
    pending_comms: List[Comm] = [comm1, comm2, comm3]

    this_actor.info("Send some data to rank-1")
    rank1_mailbox: Mailbox = Mailbox.by_name("rank1")
    for i in range(3):
        rank1_mailbox.put(i, 1)

    this_actor.info("Test for completed comms")
    while pending_comms:
        flag = Comm.test_any(pending_comms)
        if flag != -1:
            pending_comms.pop(flag)
            this_actor.info("Remove a pending comm.")
        else:
            # Nothing matches, wait for a little bit
            this_actor.sleep_for(0.1)
    this_actor.info("Last comm is complete")


def rank1():
    rank0_mailbox: Mailbox = Mailbox.by_name("rank0")
    rank1_mailbox: Mailbox = Mailbox.by_name("rank1")
    for i in range(3):
        data: int = rank1_mailbox.get()
        this_actor.info(f"Received {data}")
        msg_content = f"Message {i}"
        this_actor.info(f"Send '{msg_content}'")
        rank0_mailbox.put(msg_content, int(1e6))


def main():
    settings = create_parser().parse_known_args()[0]
    e = Engine(sys.argv)
    e.load_platform(settings.platform)

    Actor.create("rank0", e.host_by_name("Tremblay"), rank0)
    Actor.create("rank1", e.host_by_name("Fafard"), rank1)

    e.run()


if __name__ == "__main__":
    main()



#
# Copyright 2012-2015 Bronto Software, Inc. and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re

from docutils import nodes, utils
from sphinx.util.nodes import split_explicit_title

def get_javadoc_ref(app, rawtext, text):
    javadoc_url_map = app.config.javadoc_url_map

    # Add default Java SE sources
    if not javadoc_url_map.get("java"):
        javadoc_url_map["java"] = ("http://docs.oracle.com/javase/8/docs/api", 'javadoc8')
    if not javadoc_url_map.get("javax"):
        javadoc_url_map["javax"] = ("http://docs.oracle.com/javase/8/docs/api", 'javadoc8')
    if not javadoc_url_map.get("org.xml"):
        javadoc_url_map["org.xml"] = ("http://docs.oracle.com/javase/8/docs/api", 'javadoc8')
    if not javadoc_url_map.get("org.w3c"):
        javadoc_url_map["org.w3c"] = ("http://docs.oracle.com/javase/8/docs/api", 'javadoc8')

    source = None
    package = ''
    method = None

    if '(' in text:
        # If the javadoc contains a line like this:
        # {@link #sort(List)}
        # there is no package so the text.rindex will fail
        try:
            split_point = text.rindex('.', 0, text.index('('))
            method = text[split_point + 1:]
            text = text[:split_point]
        except ValueError:
            pass

    for pkg, (baseurl, ext_type) in javadoc_url_map.items():
        if text.startswith(pkg + '.') and len(pkg) > len(package):
            source = baseurl, ext_type
            package = pkg

    if not source:
        return None

    baseurl, ext_type = source

    package_parts = []
    cls_parts = []

    for part in text.split('.'):
        if cls_parts or part[0].isupper():
            cls_parts.append(part)
        else:
            package_parts.append(part)

    package = '.'.join(package_parts)
    cls = '.'.join(cls_parts)

    if not baseurl.endswith('/'):
        baseurl = baseurl + '/'

    if ext_type == 'javadoc':
        if not cls:
            cls = 'package-summary'
        source = baseurl + package.replace('.', '/') + '/' + cls + '.html'
        if method:
            source = source + '#' + method
    elif ext_type == 'javadoc8':
        if not cls:
            cls = 'package-summary'
        source = baseurl + package.replace('.', '/') + '/' + cls + '.html'
        if method:
            source = source + '#' + re.sub(r'[()]', '-', method)
    elif ext_type == 'sphinx':
        if not cls:
            cls = 'package-index'
        source = baseurl + package.replace('.', '/') + '/' + cls.replace('.', '-') + '.html'
        if method:
            source = source + '#' + package + '.' + cls + '.' + method
    else:
        raise ValueError('invalid target specifier ' + ext_type)

    title = '.'.join(filter(None, (package, cls, method)))
    node = nodes.reference(rawtext, '')
    node['refuri'] = source
    node['reftitle'] = title

    return node

def javadoc_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    """ Role for linking to external Javadoc """

    has_explicit_title, title, target = split_explicit_title(text)
    title = utils.unescape(title)
    target = utils.unescape(target)

    if not has_explicit_title:
        target = target.lstrip('~')

        if title[0] == '~':
            title = title[1:].rpartition('.')[2]

    app = inliner.document.settings.env.app
    ref = get_javadoc_ref(app, rawtext, target)

    if not ref:
         raise ValueError("no Javadoc source found for %s in javadoc_url_map" % (target,))

    ref.append(nodes.Text(title, title))

    return [ref], []
